
/* ------------------------------------------------------------------------ */
/* XCRUNCH.C (C) Copyright Bill Buckels 2013                                */
/* All Rights Reserved.                                                     */
/*                                                                          */
/* Licence Agreement                                                        */
/* -----------------                                                        */
/*                                                                          */
/* You have a royalty-free right to use, modify, reproduce and              */
/* distribute this source code in any way you find useful,                  */
/* provided that you agree that Bill Buckels has no warranty obligations    */
/* or liability resulting from said distribution in any way whatsoever.     */
/* If you don't agree, remove this source code from your computer now.      */
/*                                                                          */
/* Written by   : Bill Buckels                                              */
/* Email:         bbuckels@escape.ca                                        */
/*                                                                          */
/* Purpose      : This utility will allow you to convert from 				*/
/*                Apple II Double Hi-Res 140 x 192 x 16 color images to     */
/*				  compressed equivalent Apple II Double Hi-Res images.      */
/*                                                                          */
/*                A single purpose compression utility for A2FC and         */
/*                BIN/AUX DHGR Images for the Apple II.                     */
/*
					Outputs 2 formats to choose from:

                      XCC (XCrunch Compressed Chunk) DHGR images.
                      XRC (XCrunch Raw Chunk) DHGR images.


	The XCC is encoded with the color indices for all the pixels of
	the screen as a chunk using ZSoft PCX run length encoding.
	These indices are based on Apple II Double Low-Res Color Order.

	The XRC is the equivalent of the XCC but it is not run-length
	encoded. Therefore it will unpack faster and can be loaded more
	quickly. It could be used for busy images and monochrome images
	when run-length encoding based on 4 bit color offers little or no
	disk-space saving. The XRC is always 13445 bytes (82% or 18%
	less) when compared to the 16384 bytes of the equivalent A2FC or
	BIN/AUX BSAVED formats.

	I have not bothered to adapt this encoding to other Apple II
	Screen modes, nor to DHGR image fragments but most of this code
	could possibly be expanded for those purposes. It would be simple
	to adapt most of this logic to compressing Double Low-Res images
	like my DLO and DL1/DL2 formats but they are pretty small
	already.

	My decision to use the Double Low-Res Color Order was based on
	Color Constants that I already had in place in my Aztec C
	Graphics library routines and not some nefarious plan to confuse
	other programmers and the fact that I did not include palette
	information in my file header was to save disk space.

	The Apple II Low-Res mode (not Double Low-Res) is even less
	likely to warrant my interest as a target format for adapting
	this code, but ordinary Hi-Res mode (not Double Hi-Res) might be
	a potential target for some other utility and library routines at
	a future date. However my current thoughts on encoding indices
	for the 8 Hi-Res Colors (2 blacks, 2 whites, 4 colors) at 280
	horizontal positions (based on artifacting) are not very hopeful.
	A better method would probably be to encode a scanline level
	interlaced format based on encoded repeats of 70 bytes of 2 bit
	patterns, followed by 10 encoded bytes of monochrome data for the
	palette bits for the scanline.

	Be that as it may, this utility is what I have done at this time.

																			*/
/* Revision     : 1.0 First Release                                         */
/* ------------------------------------------------------------------------ */
/* Written in Large Model 16 bit Microsoft C (MSC) Version 8.00c            */
/* Note: Run in an MS-DOS emulator like DOSBox if you can't run it raw.     */
/* ------------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ------------------------------------------------------------------------ */
/* Declarations, Vars. etc.                                                 */
/* ------------------------------------------------------------------------ */


#define ASCIIZ 0

unsigned char *szTextTitle =
    "XCRUNCH(C) Version 1.0 Copyright Bill Buckels 2013\n"
    "All Rights Reserved.";

#define SUCCESS  0
#define VALID    SUCCESS
#define FAILURE  -1
#define INVALID  FAILURE


/* Apple 2 Double Hires Format */

/* provides base address for page1 hires scanlines  */
unsigned HB[]={
0x2000, 0x2400, 0x2800, 0x2C00, 0x3000, 0x3400, 0x3800, 0x3C00,
0x2080, 0x2480, 0x2880, 0x2C80, 0x3080, 0x3480, 0x3880, 0x3C80,
0x2100, 0x2500, 0x2900, 0x2D00, 0x3100, 0x3500, 0x3900, 0x3D00,
0x2180, 0x2580, 0x2980, 0x2D80, 0x3180, 0x3580, 0x3980, 0x3D80,
0x2200, 0x2600, 0x2A00, 0x2E00, 0x3200, 0x3600, 0x3A00, 0x3E00,
0x2280, 0x2680, 0x2A80, 0x2E80, 0x3280, 0x3680, 0x3A80, 0x3E80,
0x2300, 0x2700, 0x2B00, 0x2F00, 0x3300, 0x3700, 0x3B00, 0x3F00,
0x2380, 0x2780, 0x2B80, 0x2F80, 0x3380, 0x3780, 0x3B80, 0x3F80,
0x2028, 0x2428, 0x2828, 0x2C28, 0x3028, 0x3428, 0x3828, 0x3C28,
0x20A8, 0x24A8, 0x28A8, 0x2CA8, 0x30A8, 0x34A8, 0x38A8, 0x3CA8,
0x2128, 0x2528, 0x2928, 0x2D28, 0x3128, 0x3528, 0x3928, 0x3D28,
0x21A8, 0x25A8, 0x29A8, 0x2DA8, 0x31A8, 0x35A8, 0x39A8, 0x3DA8,
0x2228, 0x2628, 0x2A28, 0x2E28, 0x3228, 0x3628, 0x3A28, 0x3E28,
0x22A8, 0x26A8, 0x2AA8, 0x2EA8, 0x32A8, 0x36A8, 0x3AA8, 0x3EA8,
0x2328, 0x2728, 0x2B28, 0x2F28, 0x3328, 0x3728, 0x3B28, 0x3F28,
0x23A8, 0x27A8, 0x2BA8, 0x2FA8, 0x33A8, 0x37A8, 0x3BA8, 0x3FA8,
0x2050, 0x2450, 0x2850, 0x2C50, 0x3050, 0x3450, 0x3850, 0x3C50,
0x20D0, 0x24D0, 0x28D0, 0x2CD0, 0x30D0, 0x34D0, 0x38D0, 0x3CD0,
0x2150, 0x2550, 0x2950, 0x2D50, 0x3150, 0x3550, 0x3950, 0x3D50,
0x21D0, 0x25D0, 0x29D0, 0x2DD0, 0x31D0, 0x35D0, 0x39D0, 0x3DD0,
0x2250, 0x2650, 0x2A50, 0x2E50, 0x3250, 0x3650, 0x3A50, 0x3E50,
0x22D0, 0x26D0, 0x2AD0, 0x2ED0, 0x32D0, 0x36D0, 0x3AD0, 0x3ED0,
0x2350, 0x2750, 0x2B50, 0x2F50, 0x3350, 0x3750, 0x3B50, 0x3F50,
0x23D0, 0x27D0, 0x2BD0, 0x2FD0, 0x33D0, 0x37D0, 0x3BD0, 0x3FD0};


/*

The following is logically reordered to match the lores
color order...

                                                Repeated
                                                Binary
          Color         aux1  main1 aux2  main2 Pattern
          Black          00    00    00    00    0000
          Magenta        08    11    22    44    0001
		  Dark Blue      11    22    44    08    1000
          Violet         19    33    66    4C    1001
          Dark Green     22    44    08    11    0100
          Grey1          2A    55    2A    55    0101
          Medium Blue    33    66    4C    19    1100
          Light Blue     3B    77    6E    5D    1101
          Brown          44    08    11    22    0010
          Orange         4C    19    33    66    0011
          Grey2          55    2A    55    2A    1010
          Pink           5D    3B    77    6E    1011
          Green          66    4C    19    33    0110
          Yellow         6E    5D    3B    77    0111
          Aqua           77    6E    5D    3B    1110
          White          7F    7F    7F    7F    1111





*/

/*

#define LOBLACK   	0
#define LORED     	1
#define LODKBLUE 	2
#define LOPURPLE  	3
#define LODKGREEN	4
#define LOGRAY    	5
#define LOMEDBLUE	6
#define LOLTBLUE 	7
#define LOBROWN   	8
#define LOORANGE  	9
#define LOGREY    	10
#define LOPINK    	11
#define LOLTGREEN	12
#define LOYELLOW  	13
#define LOAQUA    	14
#define LOWHITE   	15

*/

/* the following array is based on the above */
unsigned char dhrbytes[16][4] = {
	0x00,0x00,0x00,0x00,
	0x08,0x11,0x22,0x44,
	0x11,0x22,0x44,0x08,
	0x19,0x33,0x66,0x4C,
	0x22,0x44,0x08,0x11,
	0x2A,0x55,0x2A,0x55,
	0x33,0x66,0x4C,0x19,
	0x3B,0x77,0x6E,0x5D,
	0x44,0x08,0x11,0x22,
	0x4C,0x19,0x33,0x66,
	0x55,0x2A,0x55,0x2A,
	0x5D,0x3B,0x77,0x6E,
	0x66,0x4C,0x19,0x33,
	0x6E,0x5D,0x3B,0x77,
	0x77,0x6E,0x5D,0x3B,
	0x7F,0x7F,0x7F,0x7F};


unsigned char dhrbuf[16384];

/* read 2 input files */
int read_binaux(unsigned char *basename)
{

	FILE *fp;
	unsigned char infile[128];

    // the bsaved images are split into two files
    // the first file is loaded into aux mem
    sprintf(infile,"%s.AUX",basename);
	fp = fopen(infile,"rb");
	if (NULL == fp)return INVALID;
	fread(dhrbuf,1,8192,fp);
	fclose(fp);

    // the second file is loaded into main mem
    sprintf(infile,"%s.BIN",basename);
	fp = fopen(infile,"rb");
	if (NULL == fp)return INVALID;
	fread(&dhrbuf[8192],1,8192,fp);
	fclose(fp);

	return SUCCESS;

}

/* read one input file */
int read_2fc(unsigned char *basename)
{
	FILE *fp;
	unsigned char infile[128];

    sprintf(infile,"%s.2FC",basename);
	fp = fopen(infile,"rb");
	if (NULL == fp)return INVALID;
	fread(dhrbuf,1,16384,fp);
	fclose(fp);

	return SUCCESS;

}


/* returns the Apple II Hires drawcolor 0-15 */
/* a double hi-res pixel can occur at any one of 7 positions */
/* in a 4 byte block which spans aux and main screen memory */
/* the horizontal resolution is 140 pixels */
int dhrgetpixel(int x,int y)
{
    int xoff, pattern, idx;
    unsigned char *ptraux, *ptrmain,c1, c2, d1, d2;

    pattern = (x%7);
	xoff = HB[y] + ((x/7) * 2);
    ptraux  = (unsigned char *) &dhrbuf[xoff-0x2000];
    ptrmain = (unsigned char *) &dhrbuf[xoff];


	switch(pattern)
	{

        /* compare colors in the input file to color patterns and return drawcolor */
 		case 0: c1 = ptraux[0] &0x0f;
		        for (idx = 0; idx < 16; idx++) {
				  d1 = dhrbytes[idx][0] & 0x0f;
				  if (d1 == c1) return idx;
				}
		        break;
		case 1: c1 = ptraux[0] & 0x70;
		        c2 = ptrmain[0] & 0x01;
		        for (idx = 0; idx < 16; idx++) {
				  d1 = dhrbytes[idx][0] & 0x70;
				  d2 = dhrbytes[idx][1] & 0x01;
				  if (d1 == c1 && d2 == c2) return idx;
				}
		        break;
		case 2: c1 = ptrmain[0] & 0x1e;
		        for (idx = 0; idx < 16; idx++) {
				  d1 = dhrbytes[idx][1] & 0x1e;
				  if (d1 == c1) return idx;
				}
		        break;
		case 3: c1 = ptrmain[0] & 0x60;
		        c2 = ptraux[1] & 0x03;
		        for (idx = 0; idx < 16; idx++) {
				  d1 = dhrbytes[idx][1] & 0x60;
				  d2 = dhrbytes[idx][2] & 0x03;
				  if (d1 == c1 && d2 == c2) return idx;
				}
                break;
		case 4: c1 = ptraux[1] & 0x3c;
		        for (idx = 0; idx < 16; idx++) {
				  d1 = dhrbytes[idx][2] & 0x3c;
				  if (d1 == c1) return idx;
				}
		        break;
		case 5: c1 = ptraux[1] & 0x40;
 		        c2 = ptrmain[1] & 0x07;
		        for (idx = 0; idx < 16; idx++) {
				  d1 = dhrbytes[idx][2] & 0x40;
				  d2 = dhrbytes[idx][3] & 0x07;
				  if (d1 == c1 && d2 == c2) return idx;
				}
                break;
		case 6: c1 = ptrmain[1] & 0x78;
		        for (idx = 0; idx < 16; idx++) {
				  d1 = dhrbytes[idx][3] & 0x78;
				  if (d1 == c1) return idx;
				}
 		        break;
	}

    return INVALID;

}


/* pcx encoder helper */
/* the writer for the encline function */
int encput(unsigned char byt,unsigned char cnt, FILE *pcxfile)
{
          if(cnt){
            if((cnt==1)&& (0xc0 != (0xc0 &byt))){
                if(EOF == fputc((int)byt,pcxfile))
                return(0);
                return(1);
            }
            else{
                if(EOF==fputc((int)0xc0|cnt,pcxfile))
                return(0);
                if(EOF==fputc((int)byt,pcxfile))
                return(0);
                return(2);
            }
        }
        return(0);

}

/* encode a line in pcxformat encoding */
/* encodes a raw line and writes it out to disk */
int encline(unsigned char *inbuff,int inlen, FILE *pcxfile)
{
    unsigned char this,last;
    int srcindex,i;
    int total;
    unsigned char runcount;
    total=0;
    last = *(inbuff); runcount=1;

for(srcindex=1;srcindex!=inlen;srcindex++){
    this= *(++inbuff);
    if(this==last){
        runcount++;
        if(runcount==63){
            if(!(i=encput(last,runcount,pcxfile)))
            return(0);
            total+=i;
            runcount=0;
        }
    }
    else{
        if(runcount){
            if(!(i=encput(last,runcount,pcxfile)))
            return(0);
            total+=i;
        }
        last=this;
        runcount=1;
      }
   }

   if(runcount){
    if(!(i=encput(last,runcount,pcxfile)))
    return(0);
    return(total+i);
    }
    return (total);

}


unsigned char bigbuf[13440];

int save_to_X(unsigned char *basename)
{

    FILE *fp, *fp2;
    unsigned char xrcfile[128], xccfile[128], c, d, byt;
	int x,x2,y,idx,jdx;

    sprintf(xccfile,"%s.XCC",basename);
    sprintf(xrcfile,"%s.XRC",basename);

	fp = fopen(xccfile,"wb");
	if (NULL == fp)return INVALID;

	fp2 = fopen(xrcfile,"wb");
	if (NULL == fp) {
		fclose(fp);
		return INVALID;
	}

    /* write a 5 byte header of sorts */

    /* some kind of identifier */
    fputc('X',fp); fputc('X',fp2);
    fputc('C',fp); fputc('R',fp2);
    fputc('C',fp); fputc('C',fp2);

    /* width in pixels x height in rasters */
    fputc(140,fp); fputc(140,fp2);
    fputc(192,fp); fputc(192,fp2);

    /* write raster oriented format rather
       than memory oriented format */
   	x2 = 0;

   	for (y = 0; y< 192; y++) {
		for (x = 0; x < 140;) {
		  idx = dhrgetpixel(x,y);
		  x++;
		  jdx = dhrgetpixel(x,y);
		  x++;
		  if (idx == INVALID || jdx == INVALID) {
			fclose(fp);
			fclose(fp2);
			remove(xccfile);
			remove(xrcfile);
			return INVALID;
		  }
		  c = (unsigned char) idx;
		  d = (unsigned char) jdx;
		  byt = (c | (d << 4));
		  bigbuf[x2] = byt;
		  x2++;
		  fputc(byt,fp2);
	   }
    }
    fclose(fp2);
    encline(bigbuf,13440,fp);
    fclose(fp);
    return SUCCESS;

}

void main(int argc, char **argv)
{
  int status = 0;

  unsigned char fname[128],infile[128],outfile[128];
  unsigned char *wordptr;

  if(argc == 1) {
    puts(szTextTitle);
    puts("Command line Usage is \"XCRUNCH MyHires.2FC\"");
    puts("                      \"XCRUNCH MyHires.BIN\"");
    puts("                      \"XCRUNCH MyHires.AUX\"");
    puts("                      \"XCRUNCH MyHires.2FC OutfileBaseName\"");
    puts("                      \"XCRUNCH MyHires.BIN OutfileBaseName\"");
    puts("                      \"XCRUNCH MyHires.AUX OutfileBaseName\"");
    puts("If converting .BIN and .AUX file pairs, both must be present.");
    printf("Enter Input FileName (Blank to Exit): ");
    gets(fname);
    if (fname[0] == ASCIIZ)
      exit(1);
    printf("Enter Output FileBaseName (Blank for Same) : ");
    gets(outfile);
  }
  else {
    strcpy(fname, argv[1]);
    if (argc > 2)
      strcpy(outfile, argv[2]);
    else
      outfile[0] = ASCIIZ;
  }

  strcpy(infile, fname);
  wordptr = strtok(infile, ".");
  if (outfile[0] == ASCIIZ)strcpy(outfile,infile);

  status = read_binaux(infile);
  if(status)status = read_2fc(infile);
  if (status) {
    puts(szTextTitle);
    printf("%s is an Unsupported Format or cannot be opened.\n", fname);
    exit(1);
  }

 status = save_to_X(outfile);
 if (status == SUCCESS)
	 printf("%s.XCC and %s.XRC Saved!\n",outfile,outfile);
 else
    printf("Error saving %s.XCC and %s.XRC!\n",outfile,outfile);

 puts("Have a Nice Dos!");
 exit(0);

}
