/* ------------------------------------------------------------------------ */
/* AWINDLO.C (C) Copyright Bill Buckels 2013                                */
/* All Rights Reserved.                                                     */
/*                                                                          */
/* Licence Agreement                                                        */
/* -----------------                                                        */
/*                                                                          */
/* You have a royalty-free right to use, modify, reproduce and              */
/* distribute this source code in any way you find useful,                  */
/* provided that you agree that Bill Buckels has no warranty obligations    */
/* or liability resulting from said distribution in any way whatsoever.     */
/* If you don't agree, remove this source code from your computer now.      */
/*                                                                          */
/* Written by   : Bill Buckels                                              */
/* Email:         bbuckels@mts.net                                          */
/*                                                                          */
/* Input Files:   This utility converts from the following input file		*/
/*                formats:													*/
/*																			*/
/* BMP Files:     560 x 384 x 256 Color BMP Files with either .BMP or		*/
/*                .256 (AWINDLO "Back-Up File) extensions, (tested with		*/
/*                Screen Captures from AppleWin 1.16, 1.17, 1.20, and 1.22.	*/
/*																			*/
/*																			*/
/* Text Files:    6 - line text files for "raw" creation of output files	*/
/*                or for layering text over other output files in either	*/
/*                DLGR or LGR format or AppleWin Screen Capture format.		*/
/*																			*/
/* Apple II DLGR: BSaved DL1 and DL2 and Raster DLO Images					*/
/*           LGR: BSaved SL2 and Raster SLO Images							*/
/*																			*/
/*  																		*/
/* Output Files:															*/
/*																			*/
/*   Default      BMP output: 160 x 128 x 24 bit Image including "palette"  */
/*   Optional                 160 x 128 x 24 bit without "palette"          */
/*                            80 x 64 x 24 bit Image including "palette"    */
/*                            80 x 48 x 24 bit without "palette"            */
/*                                                                          */
/*   Default    : Apple II DLGR: BSaved DL1 and DL2 and Raster DLO Images   */
/*   Optional               LGR: BSaved SL2 and Raster SLO Images           */
/*                                                                          */
/*                                                                          */
/*																			*/
/* Errata       : When converting from an AppleWin Screen Grab, the         */
/*                original file is saved to file with the same basename     */
/*                but with the file extension of .256. You will need to     */
/*                save these backed-up AppleWin files somewhere else        */
/*                when you are done your conversion session because any     */
/*                future "same-named" files will over-write your            */
/*                back-ups (and DLO, DL1 and DL2 files as well).            */
/*                If you don't care that's fine.                            */
/*                															*/
/*                You can also load your .256 files over and over again.	*/
/*                AWINDLO doesn't care. It just wants to keep AppleWin's	*/
/*                BMP's separate from its own.                              */
/*                                                                          */
/* Revision     : 1.0 First Release                                         */
/* ------------------------------------------------------------------------ */
/* Written in Large Model 16 bit Microsoft C (MSC) Version 8.00c            */
/* Note: Run in an MS-DOS emulator like DOSBox if you can't run it raw.     */
/* ------------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include "awindlo.h"


/* the following is more like a test for read permission,
   but if we can't open a file for reading, we likely also
   can't over-write it, so everything else is likely to fail
   right down the line... */
int FileExists(char *fname)
{
	FILE *fp=NULL;

	fp = fopen(fname, "rb");
	if (NULL == fp) return 0;
	fclose(fp);

	return 1;
}


/* back-up previous sessions if any with the same basename.
advise the user what we are doing so so they can move or rename
the back-up if they need to before it is overwitten again.
*/

/* if we are converting a bmp from AppleWin, the original
will be backed-up to basename.256 during processing.

if we are converting from any of our other input formats and
we find a bmp of the same name, back it up to basename.DIB
or it will be overwritten.

use the same sort of immediate logic for our other files... */

int backup = 0, silent = 0;
void BackUpOldFile(char *basename, char *src, char *dest)
{

	char srcname[128], destname[128];

	sprintf(srcname,"%s.%s",basename,src);
	sprintf(destname,"%s.%s",basename,dest);

	if (FileExists(srcname)) {

		if (FileExists(destname)) {
		   backup++;
		   remove(destname);
		   if (silent == 0)printf("Abandoned File %s Deleted!\n",destname);
		}

		rename(srcname,destname);
		if (silent == 0)printf("%s Renamed to %s.\n",srcname,destname);
	}

}

/* Bitmap Header structures */
typedef struct tagRGBQUAD
{
    unsigned char    rgbBlue;
    unsigned char    rgbGreen;
    unsigned char    rgbRed;
    unsigned char    rgbReserved;
} RGBQUAD;

typedef struct tagBITMAPINFOHEADER
{
    unsigned long   biSize;
    unsigned long   biWidth;
    unsigned long   biHeight;
    unsigned int    biPlanes;
    unsigned int    biBitCount;
    unsigned long   biCompression;
    unsigned long   biSizeImage;
    unsigned long   biXPelsPerMeter;
    unsigned long   biYPelsPerMeter;
    unsigned long   biClrUsed;
    unsigned long   biClrImportant;
} BITMAPINFOHEADER;

/* constant for the biCompression field */
#define BI_RGB      0L

typedef struct tagBITMAPFILEHEADER
{
    unsigned char   bfType[2];
    unsigned long   bfSize;
    unsigned int    bfReserved1;
    unsigned int    bfReserved2;
    unsigned long   bfOffBits;
} BITMAPFILEHEADER;


typedef struct tagBMPHEADER
{
BITMAPFILEHEADER bfi;
BITMAPINFOHEADER bmi;
} BMPHEADER;

/* static structures for processing input file */
BITMAPFILEHEADER BitMapFileHeader;
RGBQUAD          sbmp[256]; /* super vga - applewin */
BITMAPINFOHEADER bmi;
BMPHEADER mybmp;


/* header for 160 x 128 x 24 bit BMP output */
unsigned char BMP_header[] ={
0x42, 0x4D, 0x36, 0xF0, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x36, 0x00, 0x00, 0x00, 0x28, 0x00,
0x00, 0x00, 0xA0, 0x00, 0x00, 0x00, 0x80, 0x00,
0x00, 0x00, 0x01, 0x00, 0x18, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0xF0, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00};


/* HColors used for applewin version detection */
/* applewin rendering - older version - 1.16 */
/* in applewin color order */
unsigned char cmappleold[6][3] = {
  0,  0,  0,
255,255,254,
  0,128,255,
240, 80,  0,
 32,192,  0,
160,  0,255};

/* applewin rendering - newer versions - 1.17? to 1.22 */
unsigned char cmapplenew[6][3] = {
  0,  0,  0,
255,255,254,
 13,161,255,
242, 94,  0,
 56,203,  0,
199, 52,255};

unsigned char *szTextTitle =
    "AWINDLO(C) Version 1.0 Copyright Bill Buckels 2013\n"
    "All Rights Reserved.";

#define ASCIIZ     '\x00'
#define SUCCESS  0
#define INVALID  -1

#define LOBLACK   	0
#define LORED     	1
#define LODKBLUE 	2
#define LOPURPLE  	3
#define LODKGREEN	4
#define LOGRAY    	5
#define LOMEDBLUE	6
#define LOLTBLUE 	7
#define LOBROWN   	8
#define LOORANGE  	9
#define LOGREY    	10
#define LOPINK    	11
#define LOLTGREEN	12
#define LOYELLOW  	13
#define LOAQUA    	14
#define LOWHITE   	15


#define NUM_RGB_COLORS 3

enum {  BLACK = 0,
        BLUE,
        GREEN,
        CYAN,
        RED,
        MAGENTA,
        BROWN,
        GRAY,
        WHITE,
        LBLUE,
        LGREEN,
        LCYAN,
        LRED,
        LMAGENTA,
        YELLOW,
        BWHITE,
        NUM_VGA_COLORS};

unsigned char VGAtoApple[NUM_VGA_COLORS]={
	LOBLACK,
	LODKBLUE,
	LODKGREEN,
	LOMEDBLUE,
	LORED,
	LOPURPLE,
	LOBROWN,
	LOGRAY,
	LOWHITE,
	LOLTBLUE,
	LOLTGREEN,
	LOAQUA,
	LOORANGE,
	LOPINK,
	LOYELLOW,
	LOWHITE};

/* the following 4 palettes are used to troll
   for standard colors in the order given below */

/* 16 Color BMP style palette (probably PBRUSH from Windows 3.1) */
unsigned char rgbBmpArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
0x00, 0x00, 0x00,    // BLACK
0x00, 0x00, 0xBF,    // BLUE
0x00, 0xBF, 0x00,    // GREEN
0x00, 0xBF, 0xBF,    // CYAN
0xBF, 0x00, 0x00,    // RED
0xBF, 0x00, 0xBF,    // MAGENTA
0xBF, 0xBF, 0x00,    // BROWN
0x80, 0x80, 0x80,    // GRAY
0xC0, 0xC0, 0xC0,    // WHITE
0x00, 0x00, 0xFF,    // LBLUE
0x00, 0xFF, 0x00,    // LGREEN
0x00, 0xFF, 0xFF,    // LCYAN
0xFF, 0x00, 0x00,    // LRED
0xFF, 0x00, 0xFF,    // LMAGENTA
0xFF, 0xFF, 0x00,    // YELLOW
0xFF, 0xFF, 0xFF};   // BWHITE

/* 16 Color BMP style palette (Windows XP) */
/* notice that these johnny-come-latelies reversed GRAY and WHITE */
unsigned char rgbXmpArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
0x00, 0x00, 0x00,  	 // BLACK
0x00, 0x00, 0x80,  	 // BLUE
0x00, 0x80, 0x00,  	 // GREEN
0x00, 0x80, 0x80,  	 // CYAN
0x80, 0x00, 0x00,  	 // RED
0x80, 0x00, 0x80,  	 // MAGENTA
0x80, 0x80, 0x00,  	 // BROWN
0x80, 0x80, 0x80,  	 // GRAY
0xC0, 0xC0, 0xC0,  	 // WHITE
0x00, 0x00, 0xFF,  	 // LBLUE
0x00, 0xFF, 0x00,  	 // LGREEN
0x00, 0xFF, 0xFF,  	 // LCYAN
0xFF, 0x00, 0x00,  	 // LRED
0xFF, 0x00, 0xFF,  	 // LMAGENTA
0xFF, 0xFF, 0x00,  	 // YELLOW
0xFF, 0xFF, 0xFF};	 // BWHITE

// VGA Palette
unsigned char rgbVgaArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
0x00, 0x00, 0x00,            /* BLACK    */
0x00, 0x00, 0xFF,            /* BLUE     */
0x00, 0xFF, 0x00,            /* GREEN    */
0x00, 0xFF, 0xFF,            /* CYAN     */
0xFF, 0x00, 0x00,            /* RED      */
0xFF, 0x00, 0xFF,            /* MAGENTA  */
0xFF, 0xFF, 0x00,            /* BROWN    */
0x55, 0x55, 0x55,            /* GRAY     */
0xC0, 0xC0, 0xC0,            /* WHITE    */
0x55, 0x55, 0xFF,            /* LBLUE    */
0x55, 0xFF, 0x55,            /* LGREEN   */
0x55, 0xFF, 0xFF,            /* LCYAN    */
0xFF, 0x55, 0x55,            /* LRED     */
0xFF, 0x55, 0xFF,            /* LMAGENTA */
0xFF, 0xFF, 0x55,            /* YELLOW   */
0xFF, 0xFF, 0xFF};           /* BWHITE   */

// 16 color ZSOFT PCPAINT PCX style palette
unsigned char rgbPcxArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
0x00, 0x00, 0x00,            /* BLACK    */
0x00, 0x00, 0xAA,            /* BLUE     */
0x00, 0xAA, 0x00,            /* GREEN    */
0x00, 0xAA, 0xAA,            /* CYAN     */
0xAA, 0x00, 0x00,            /* RED      */
0xAA, 0x00, 0xAA,            /* MAGENTA  */
0xAA, 0xAA, 0x00,            /* BROWN    */
0x55, 0x55, 0x55,            /* GRAY     */
0xAA, 0xAA, 0xAA,            /* WHITE    */
0x55, 0x55, 0xFF,            /* LBLUE    */
0x55, 0xFF, 0x55,            /* LGREEN   */
0x55, 0xFF, 0xFF,            /* LCYAN    */
0xFF, 0x55, 0x55,            /* LRED     */
0xFF, 0x55, 0xFF,            /* LMAGENTA */
0xFF, 0xFF, 0x55,            /* YELLOW   */
0xFF, 0xFF, 0xFF};           /* BWHITE   */

/* the following attempts to match BMP colors with a common palette */
/* that is to say... a basic default EGA style 16 color palette like */
/* the kind that Windows Paint provides as a lowest common denominator */
/* my rationale here is that if we can't automate the entire remapping */
/* we fail the whole sh*terree because otherwise we could start mashing */
/* two colors or more together and I for one would not care for that... */
int GetVGAIndex(unsigned char red, unsigned char green, unsigned char blue)
{

    int idx;

    // try exact match
    for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
		if (red == rgbBmpArray[idx][0] && green == rgbBmpArray[idx][1] && blue == rgbBmpArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
		if (red == rgbXmpArray[idx][0] && green == rgbXmpArray[idx][1] && blue == rgbXmpArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
        if (red == rgbVgaArray[idx][0] && green == rgbVgaArray[idx][1] && blue == rgbVgaArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
        if (red == rgbPcxArray[idx][0] && green == rgbPcxArray[idx][1] && blue == rgbPcxArray[idx][2])return idx;
	}

	// adjust gun values to match using EGA-like thresholds
	// this corresponds to the old PCX style palette
	if (red < 43) red = 0;         // 0x00
	else if (red < 128) red = 85;  // 0x55
	else if (red < 224) red = 170; // 0xaa
	else red = 255;                // 0xff

	if (green < 43) green = 0;
	else if (green < 128) green = 85;
	else if (green < 213) green = 170;
	else green = 255;

	if (blue < 43) blue = 0;
	else if (blue < 128) blue = 85;
	else if (blue < 213) blue = 170;
	else blue = 255;

    // try again
    for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
		if (red == rgbBmpArray[idx][0] && green == rgbBmpArray[idx][1] && blue == rgbBmpArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
		if (red == rgbXmpArray[idx][0] && green == rgbXmpArray[idx][1] && blue == rgbXmpArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
        if (red == rgbVgaArray[idx][0] && green == rgbVgaArray[idx][1] && blue == rgbVgaArray[idx][2])return idx;
	}
	for (idx = 0; idx < NUM_VGA_COLORS; idx++) {
        if (red == rgbPcxArray[idx][0] && green == rgbPcxArray[idx][1] && blue == rgbPcxArray[idx][2])return idx;
	}
	// if no match yet let them pick lores color manually
    return INVALID;
}

/* this palette is in the actual apple II lores color order */
/* this shows us a[[roximately what we will end-up with */
unsigned char rgbLoresArray[NUM_VGA_COLORS][NUM_RGB_COLORS]={
  0,   0,   0,	/* black    */
227,  30,  96,	/* red      */
 96,  78, 189,	/* dk blue  */
255,  68, 253,	/* purple   */
  0, 163,  96,	/* dk green */
156, 156, 156,	/* gray     */
 20, 207, 253,	/* med blue */
208, 195, 255,	/* lt blue  */
 96, 114,   3,	/* brown    */
255, 106,  60, 	/* orange   */
156, 156, 156,	/* grey     */
255, 160, 208,	/* pink     */
 20, 245,  60,	/* lt green */
208, 221, 141,	/* yellow   */
114, 255, 208, 	/* aqua     */
255, 255, 255};	/* white    */

unsigned char awinoldcolors[16][4] = {
	0,0,0,      LOBLACK,    /* black    */
	208,0,48,	LORED,      /* red      */
	0,0,128,	LODKBLUE,   /* dk blue  */
	255,0,255,	LOPURPLE,   /* purple   */
	0,128,0,	LODKGREEN,  /* dk green */
	128,128,128,LOGRAY,     /* gray     */
	0,0,255,	LOMEDBLUE,  /* med blue */
	96,160,255,	LOLTBLUE,   /* lt blue  */
	128,80,0,	LOBROWN,    /* brown    */
    255,128,0,	LOORANGE,   /* orange   */
    192,192,192,LOGREY,     /* grey     */
    255,144,128,LOPINK,     /* pink     */
    0,255,0,	LOLTGREEN,  /* lt green */
    255,255,0,	LOYELLOW,   /* yellow   */
    64,255,144,	LOAQUA,     /* aqua     */
    255,255,255,LOWHITE};   /* white    */

unsigned char awinnewcolors[16][4] = {
	0,0,0,      LOBLACK,    /* black    */
	157,9,102,  LORED,      /* red      */
	42,42,229,  LODKBLUE,   /* dk blue  */
	199,52,255, LOPURPLE,   /* purple   */
	0,118,26,   LODKGREEN,  /* dk green */
	128,128,128,LOGRAY,     /* gray     */
	13,161,255, LOMEDBLUE,  /* med blue */
	170,170,255,LOLTBLUE,   /* lt blue  */
	85,85,0,    LOBROWN,    /* brown    */
    242,94,0,   LOORANGE,   /* orange   */
    192,192,192,LOGREY,     /* grey     */
    255,137,229,LOPINK,     /* pink     */
    56,203,0,   LOLTGREEN,  /* lt green */
    213,213,26, LOYELLOW,   /* yellow   */
    98,246,153, LOAQUA,     /* aqua     */
    255,255,254,LOWHITE};   /* white    */

/* the following is used to remap
   double lo res 4 bit colors
   from bank 0 to bank 1 */
unsigned char dloauxcolor[16] = {
	0,8,1,9,2,10,3,11,4,12,5,13,6,14,7,15};

/* the following is used to remap
   double lo res 4 bit colors
   from bank 1 to bank 0 */
unsigned char dlomaincolor[16] = {
	0,2,4,6,8,10,12,14,1,3,5,7,9,11,13,15};

/* applewin version over-ride for future use, see notes below */
unsigned int applewinversion = 0, outputscale = 2, outputpalette = 1, bmp256 = 0, lores = 0;

unsigned int fontpitch = 7, fontcolor = 15, fontbackground = 0, fontjustify = 1, fontfix = 1,
             textoverlay = 0, rainbowtext = 0, border = 255, stripedtext = 0, candycanetext = 0,
             fontccolor = LOORANGE, fontcbg=LOWHITE;

unsigned char rainbow[32] = { 7,9,10,11,12,13,14,15,
                              7,9,10,11,12,13,14,15,
                              7,9,10,11,12,13,14,15,
                              7,9,10,11,12,13,14,15};

/* the following routine attempts to match double lo-res colors from various versions of
   double lo-res AppleWin screen captures...
   until AppleWin version 1.17 auxiliary colors for double lo res were broken and must be remapped. */

unsigned char GetDLGRColor(unsigned char red, unsigned char green, unsigned char blue, int bank0)
{
	int idx, ch;

    if (applewinversion == 116) {
		/* remap auxiliary colors if version 1.16 and prior */
		for (idx = 0; idx < 16; idx++) {
			if (red == awinoldcolors[idx][0] && green == awinoldcolors[idx][1] && blue == awinoldcolors[idx][2]) {
			   ch = awinoldcolors[idx][3];
			   /* remap incorrect AppleWin 1.16 screen capture color values from bank 1
			      see notes above */
			   if (bank0 == 0) ch = dlomaincolor[ch];
			   return ch;
		    }
		}
		return 0;
	}

    /* default to 1.22 colors for now */
	for (idx = 0; idx < 16; idx++) {
		if (red == awinnewcolors[idx][0] && green == awinnewcolors[idx][1] && blue == awinnewcolors[idx][2])
		  return awinnewcolors[idx][3];
	}

    /* from here color mapping gets expensive, but the user should have
       read the documentation and source code first instead of trying some
       wierd colors that this program doesn't recognize... maybe they're just trying
       to break this program in which case who cares!

       the other side of this of course is if AppleWin decides to change its
       double lo-res colors in some future version, at which point we would have
       no way of knowing what those may change to, and in fact if the 256 color format
       for screen grabs will stay within current guidelines or even be used at all.

       the other thing of course is if the user has grabbed screens from some other
       emulator in 256 color bmp format of the same size as AppleWin's screen captures.
       probably unlikely but who knows...

       */


	/* try 1.16 colors next */
	for (idx = 0; idx < 16; idx++) {
		if (red == awinoldcolors[idx][0] && green == awinoldcolors[idx][1] && blue == awinoldcolors[idx][2])
		  return awinoldcolors[idx][3];
	}

	/* try BMP2LO colors next */
	for (idx = 0; idx < 16; idx++) {
		if (red == rgbLoresArray[idx][0] && green == rgbLoresArray[idx][1] && blue == rgbLoresArray[idx][2])
		  return (unsigned char)idx;
	}

	/* finally try for an EGA style 16 color standard palette match...
	   otherwise it's some other wierd color so
	   paint it black. */

	idx = GetVGAIndex(red,green,blue);
	if (idx > 0) return VGAtoApple[idx];
	return 0;
}



/* routines to save to Apple 2 Lores Format */

/* base addresses for Apple II primary text page are */
/* also the base addresses for the 48 scanline pairs */
/* for Apple II lores graphics mode 40 x 48 x 16 colors */
int textbase[24]={
    0x0400,
    0x0480,
    0x0500,
    0x0580,
    0x0600,
    0x0680,
    0x0700,
    0x0780,
    0x0428,
    0x04A8,
    0x0528,
    0x05A8,
    0x0628,
    0x06A8,
    0x0728,
    0x07A8,
    0x0450,
    0x04D0,
    0x0550,
    0x05D0,
    0x0650,
    0x06D0,
    0x0750,
    0x07D0};

#define RAGWIDTH  80
#define RAGHEIGHT 48
#define RAGSIZE   1920
#define BINSIZE   1016
unsigned char lobuf[RAGSIZE];
unsigned char bmpbuf[3840];

/* gets the pixel in the bmpbuf */
unsigned char getpixel(int x, int y)
{
	unsigned char pixel;
	int offset;

	if (x < 0 || x > 79) x = 0;
	if (y < 0 || y > 47) y = 0;

	offset = (y * 80) + x;
	pixel = bmpbuf[offset];

	return pixel;

}

/* sets the pixel in the bmpbuf */
void setpixel(int x, int y, unsigned char pixel)
{
	int offset;

	if (x < 0 || x > 79) return;
	if (y < 0 || y > 47) return;
	if (pixel > 15) pixel  = 0;

	offset = (y * 80) + x;
	bmpbuf[offset] = pixel;
}

/* sets the pixels in the lores buffer (lobuf) */
void setlopixel(unsigned char color,int x, int y,int ragflag)
{
     unsigned char *crt, c1, c2;
     int y1, offset;

     y1 = y / 2;
     c2 = (unsigned char ) (color & 15);

     if (y%2 == 0) {
		 /* even rows in low nibble */
		 /* mask value to preserve high nibble */
		 c1 = 240;
	 }
	 else {
		 /* odd rows in high nibble */
		 /* mask value to preserve low nibble */
		 c1 = 15;
		 c2 = c2 * 16;
	 }

     if (ragflag)
		 offset = (y1 * 80) + x;
     else
		 offset = (textbase[y1]-1024)+x;

	 crt = (unsigned char *)&lobuf[offset];
     crt[0] &= c1;
     crt[0] |= c2;
}

/* gets the pixels in the lores buffer (lobuf) */
unsigned char getlopixel(int x, int y,int ragflag)
{
     unsigned char *crt, color;
     int y1, offset;

     y1 = y / 2;

     if (ragflag)
		 offset = (y1 * 80) + x;
     else
		 offset = (textbase[y1]-1024)+x;

	 crt = (unsigned char *)&lobuf[offset];

     if (y%2 == 0) {
		 /* even rows in low nibble */
		 /* mask value to preserve high nibble */
		 color = (*crt & 0xf);
	 }
	 else {
		 /* odd rows in high nibble */
		 /* mask value to preserve low nibble */
		 color = (*crt >> 4) &0xf;
	 }
	 return color;
}


/* save 2 types of Apple II output files */
int savelofragment(unsigned char *basename)
{

	FILE *fp;
	unsigned char outfile[128], temp, remap;
	int x,y,x2,offset;

    /* save single lo-res and double lo-res */
    if (lores == 1) {
		sprintf(outfile,"%s.SLO",basename);
	}
	else {
		BackUpOldFile(basename, "DLO", "BLO");
		sprintf(outfile,"%s.DLO",basename);
	}

	fp = fopen(outfile,"wb");
	if (NULL == fp)return INVALID;

    // On the double lo res display each byte in
    // high memory is interleaved with a byte in low memory
   	memset(lobuf,0,RAGSIZE);
	for (y = 0; y< 48; y++) {
		// first 40 bytes goes to auxiliary memory (even pixels)
		for (x = 0; x < 40; x++) {
			x2 = (x*2);
			temp = getpixel(x2,y);
			remap = dloauxcolor[temp];
			setlopixel(remap,x,y,1);
		}
		// followed by the interleaf (odd pixels)
		// next 40 bytes goes to main memory
		for (x = 0; x < 40; x++) {
			x2 = (x*2) + 1;
			temp = getpixel(x2,y);
			setlopixel(temp,x+40,y,1);
		}
	}

	if (lores == 1) {
	    fputc(40,fp); // bytes
        fputc(24,fp);
	    for (y = 0; y < 24; y++) {
 	 		offset = (y * 80)+40;
	 		fwrite((unsigned char *)&lobuf[offset],1,40,fp);
		}
	}
	else {
		fputc(80,fp); // bytes
		fputc(24,fp); // bytes (rasters / 2)
		fwrite(lobuf,1,RAGSIZE,fp);
	}
	fclose(fp);

    if (lores == 0) {
   		// for double lo-res */
   		// the bsaved images are split into two files
    	// the first file is loaded into aux mem
		BackUpOldFile(basename, "DL1", "BL1");
		sprintf(outfile,"%s.DL1",basename);
		fp = fopen(outfile,"wb");
		if (NULL == fp)return INVALID;

		memset(lobuf,0,BINSIZE);
		for (y = 0; y< 48; y++) {
			for (x = 0; x < 40; x++) {
				x2 = (x*2);
				temp = getpixel(x2,y);
				remap = dloauxcolor[temp];
				setlopixel(remap,x,y,0);
			}
		}
		fwrite(lobuf,1,BINSIZE,fp);
		fclose(fp);
	}

    // for single lo res only 1 file is needed
    // for double lo res
    // the second file is loaded into main mem
    if (lores == 1)
    	sprintf(outfile,"%s.SL2",basename);
    else {
		BackUpOldFile(basename, "DL2", "BL2");
    	sprintf(outfile,"%s.DL2",basename);
	}

	fp = fopen(outfile,"wb");
	if (NULL == fp)return INVALID;
	memset(lobuf,0,BINSIZE);
	for (y = 0; y< 48; y++) {
		for (x = 0; x < 40; x++) {
			x2 = (x*2) + 1;
			temp = getpixel(x2,y);
			setlopixel(temp,x,y,0);
		}
	}
	fwrite(lobuf,1,BINSIZE,fp);
	fclose(fp);

	return SUCCESS;
}



unsigned char bmpscanline[560];

int save_to_bmp24(char *basename)
{

    FILE *fp;
    unsigned char outfile[128], tempr, tempg, tempb;
	int x,y,y2,idx,jdx,kdx;
	unsigned width, height,packet;

    sprintf(outfile,"%s.BMP",basename);

	fp = fopen(outfile,"wb");
	if (NULL == fp)return INVALID;


    memset((char *)&mybmp.bfi.bfType,0,sizeof(BMPHEADER));
    memcpy((char *)&mybmp.bfi.bfType,(unsigned char *)&BMP_header[0],sizeof(BMPHEADER));


    mybmp.bmi.biSizeImage = 0L;
    if (outputscale == 2) {

		/* write double pixels and double scanlines */
		/* write header for 160 x 128 x 24 bit bmp */
		/* if the outputpalette flag is set...
		/* bottom 32 lines are palette data */
		/* top 96 lines are image data */

		packet = 480;
		width = 160;
		if (outputpalette == 1) height = 128;
		else height = 96;
	}
	else {
		packet = 240;
		width = 80;
		if (outputpalette == 1) height = 64;
		else height = 48;
	}

    mybmp.bmi.biWidth  = (unsigned long) width;
    mybmp.bmi.biHeight = (unsigned long) height;

    fwrite((char *)&mybmp.bfi.bfType,sizeof(BMPHEADER),1,fp);

    if (outputpalette == 1) {

		/* write rgb triples */
		/* write 32 line visible palette of double lines first */
		/* write black line at bottom to white line at top */
		for (idx = 0; idx < 16; idx++) {
		   tempr = awinnewcolors[idx][0];
		   tempg = awinnewcolors[idx][1];
		   tempb = awinnewcolors[idx][2];
		   for (jdx = 0,kdx =0; jdx < width; jdx++) {
			 /* reverse order */
			 bmpscanline[kdx] = tempb; kdx++;
			 bmpscanline[kdx] = tempg; kdx++;
			 bmpscanline[kdx] = tempr; kdx++;
		   }
		   fwrite((char *)&bmpscanline[0],1,packet,fp);
		   if (outputscale == 2)fwrite((char *)&bmpscanline[0],1,packet,fp);
		}
	}


    y2 = 47;
   	for (y = 0; y< 48; y++) {

	   for (x = 0, kdx = 0; x < 80; x++) {
		  idx = (int)getpixel(x,y2);

          // range check
          if (idx < 0 || idx > 15)idx = 0; // default black

 	   	  tempr = awinnewcolors[idx][0];
	   	  tempg = awinnewcolors[idx][1];
	      tempb = awinnewcolors[idx][2];

		  bmpscanline[kdx] = tempb; kdx++;
		  bmpscanline[kdx] = tempg; kdx++;
		  bmpscanline[kdx] = tempr; kdx++;

          if (outputscale == 2) {
				bmpscanline[kdx] = tempb; kdx++;
		  		bmpscanline[kdx] = tempg; kdx++;
		  		bmpscanline[kdx] = tempr; kdx++;
		  }
	   }
	   fwrite((char *)&bmpscanline[0],1,packet,fp);
	   if (outputscale == 2) fwrite((char *)&bmpscanline[0],1,packet,fp);

	   y2 -= 1;
    }

    fclose(fp);

    return SUCCESS;

}

int AppleWinRead(unsigned char *basename)
{

 	FILE *fp;
    int x, x1, y, i, j, bank0, packet=INVALID;
    unsigned char bmpfile[128],backup[128], ch, pixel;

    if (bmp256 == 0)
    	sprintf(bmpfile,"%s.BMP",basename);
    else
		sprintf(bmpfile,"%s.256",basename);

    if((fp=fopen(bmpfile,"rb"))==NULL) {
		return INVALID;
	}

    /* read the header stuff into the appropriate structures */
    fread((char *)&BitMapFileHeader.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&bmi.biSize,
                 sizeof(BITMAPINFOHEADER),1,fp);

    if (bmi.biCompression==BI_RGB &&
        BitMapFileHeader.bfType[0] == 'B' && BitMapFileHeader.bfType[1] == 'M' &&
        bmi.biPlanes==1 && bmi.biBitCount == 8 && bmi.biWidth == 560 && bmi.biHeight == 384) {

        packet = (int) bmi.biWidth;

		fread((char *)&sbmp[0].rgbBlue, sizeof(RGBQUAD)*256,1,fp);
		/* check for AppleWin 256 color palette collapse correction HCOLOR = HWHITE
		   index 17 */
		if (sbmp[17].rgbRed == 255 && sbmp[17].rgbGreen == 255 && sbmp[17].rgbBlue == 254) {

           /* for the version over-ride assume future versions will
              use the same double lo-res colors as version 1.2 */

           if (applewinversion == 0) {
				/* if no version over-ride, check applewin version */
				/* use hcolors to compare */
				for (i = 16,j=0; j < 6; i++,j++) {
					 applewinversion = 0;
					 if (sbmp[i].rgbRed != cmapplenew[j][0]) break;
					 if (sbmp[i].rgbGreen != cmapplenew[j][1]) break;
					 if (sbmp[i].rgbBlue != cmapplenew[j][2]) break;
					 applewinversion = 120;
				}
				if (applewinversion == 0) {
					for (i = 16,j=0; j < 6; i++,j++) {
						 applewinversion = 0;
						 if (sbmp[i].rgbRed != cmappleold[j][0]) break;
						 if (sbmp[i].rgbGreen != cmappleold[j][1]) break;
						 if (sbmp[i].rgbBlue != cmappleold[j][2]) break;
						 applewinversion = 117;
					}
				}
			}
		}
  	}

    if(packet == INVALID){
		fclose(fp);
		return INVALID;
	}

    if (applewinversion == 116 || applewinversion == 117 || applewinversion == 120)
    	printf("Using Color Mapping for AppleWin Version %d.\n",applewinversion);

	memset(&bmpbuf[0],0,3840); /* clear write buffer */

    for(i=48,y=47;i>0;i--,y--)
    {
	  /* process every 8th line */
	  for (j = 0; j < 8; j++)
	  	fread((char *)&bmpscanline[0],1,packet,fp);

	  for (x = 0, x1=0; x < 80; x++, x1+=7) {
	     ch = bmpscanline[x1];
	     bank0 = (x%2);
         pixel = GetDLGRColor(sbmp[ch].rgbRed, sbmp[ch].rgbGreen, sbmp[ch].rgbBlue, bank0);
	     setpixel(x,y,pixel);

	  }
    }
    fclose(fp);

    /* we are done reading */
    /* rename the applewin file unless we are re-reading a 256 file */
    if (bmp256 == 0)BackUpOldFile(basename, "BMP", "256");

return SUCCESS;

}

/* single lo-res pixel translator */
unsigned char getslopixel(int x, int y,int ragflag)
{
     unsigned char *crt, color;
     int y1, offset;

     y1 = y / 2;

     if (ragflag)
		 offset = (y1 * 40) + x;
     else
		 offset = (textbase[y1]-1024)+x;

	 crt = (unsigned char *)&lobuf[offset];

     if (y%2 == 0) {
		 /* even rows in low nibble */
		 /* mask value to preserve high nibble */
		 color = (*crt & 0xf);
	 }
	 else {
		 /* odd rows in high nibble */
		 /* mask value to preserve low nibble */
		 color = (*crt >> 4) &0xf;
	 }
	 return color;
}



/* read Apple II DLGR and L-Res files */
 int DLO_Read(unsigned char *basename, unsigned char *fullname)
{

	FILE *fp, *fp2;
	unsigned char infile[128], temp, remap, d,l,o;
	int x,y,x2, width, height;

	y = strlen(fullname);
	if (y < 5)return INVALID;
	o = fullname[y];
	y-=4;
	d = toupper(fullname[y+1]);
	l = toupper(fullname[y+2]);
	if (fullname[y] != '.' || (d != 'D' && d!='S') || (l != 'L' && l != 'T') || o != ASCIIZ) return INVALID;
	o = toupper(fullname[y+3]);

    memset(&bmpbuf[0],0,3840); /* clear write buffer */

	if (o == 'O') {
		/* read a raster file */
		if (d == 'D') {
			if (l == 'L') sprintf(infile,"%s.DLO",basename);
			else sprintf(infile,"%s.DTO",basename);
		}
		else {
			if (l == 'L') sprintf(infile,"%s.SLO",basename);
			else sprintf(infile,"%s.STO",basename);
		}
		fp = fopen(infile,"rb");
		if (NULL != fp) {

			/* verify in case someone is mucking about with
			   renamed files - a very minimalistic precaution
			   generally speaking we don't worry what they feed us
			   as long as we don't crash... we're not flying to the
			   moon in this program so no harm done anyway. */
			width = fgetc(fp);
			height = fgetc(fp);

			if (width == 80 && (height == 24 || height == 20)) {
				// On the double lo res display each byte in
				// high memory is interleaved with a byte in low memory
				memset(lobuf,0,RAGSIZE);
				fread(lobuf,1,RAGSIZE,fp);
				fclose(fp);
				for (y = 0; y< 48; y++) {
					if (y > 39 && height == 20) break;
					// first 40 bytes goes to auxiliary memory (even pixels)
					for (x = 0; x < 40; x++) {
						x2 = (x*2);
						temp = getlopixel(x,y,1);
						remap = dlomaincolor[temp];
						setpixel(x2,y,remap);
					}
					// followed by the interleaf (odd pixels)
					// next 40 bytes goes to main memory
					for (x = 0; x < 40; x++) {
						x2 = (x*2) + 1;
						temp = getlopixel(x+40,y,1);
						setpixel(x2,y,temp);
					}
				}
				return SUCCESS;
			}
            /* single lo-res */
			if (width == 40 && (height == 24 || height == 20)) {
				memset(lobuf,0,RAGSIZE);
				fread(lobuf,1,RAGSIZE/2,fp);
				fclose(fp);
				for (y = 0; y< 48; y++) {
					if (y > 39 && height == 20) break;
					for (x = 0; x < 40; x++) {
						x2 = (x*2);
						temp = getslopixel(x,y,1);
						/* fake-out double lo-res */
						setpixel(x2,y,temp);
						setpixel(x2+1,y,temp);
					}
				}
				return SUCCESS;
			}
			fclose(fp);
			return INVALID;
		}
	}



    if (d == 'D') {
		// for double lo-res only
		// the bsaved images are split into two files
		// the first file is loaded into aux mem
		// the second file is loaded into main mem
		// both files must exist or we fail

		// for single lo-res we only open one file

		sprintf(infile,"%s.DL1",basename);
		fp = fopen(infile,"rb");
		if (NULL == fp)return INVALID;
	}


    /* for double lo-res
       leave dl2 open while we read the dl1
       we just needed to make sure both files
       are present

       for single lo-res we only open one file

       */
    if (d == 'D')
    	sprintf(infile,"%s.DL2",basename);
    else
    	sprintf(infile,"%s.SL2",basename);

	fp2 = fopen(infile,"rb");
	if (NULL == fp2) {
		if (d == 'D') fclose(fp);
		return INVALID;
	}

    /* double hi-res only
       for single lo-res we only open one file...
       not this one because it needs to be translated
       from aux colors to main colors */
    if (d == 'D') {
		memset(lobuf,0,BINSIZE);
		fread(lobuf,1,BINSIZE,fp);
		fclose(fp);
		for (y = 0; y< 48; y++) {
			for (x = 0; x < 40; x++) {
				x2 = (x*2);
				temp = getlopixel(x,y,0);
				remap = dlomaincolor[temp];
				setpixel(x2,y,remap);
			}
		}
	}

    /* finally read the dl2 or the sl2 */
	/* they have exactly the same colors
	   so for our bmp buffer we just fill
	   with double pixels to fake-out
	   the DL1 aux file */
	memset(lobuf,0,BINSIZE);
	fread(lobuf,1,BINSIZE,fp2);
	fclose(fp2);
	for (y = 0; y< 48; y++) {
		for (x = 0; x < 40; x++) {
			x2 = (x*2);
			temp = getlopixel(x,y,0);
			if (d == 'S')setpixel(x2,y,temp);
			setpixel(x2+1,y,temp);

		}
	}

	return SUCCESS;
}


unsigned int rowjustify[6];
unsigned int rowpitch[6];
unsigned int rowfix[6];
unsigned int rowcolor[6];
unsigned int rowbackground[6];
unsigned int rainbowrow[6];
unsigned int stripedrow[6];
unsigned int candycanerow[6];
unsigned int rowccolor[6];
unsigned int rowcbg[6];

/* sets row-level font controls */
/* think globally, act rowbowly */
void initFontRows()
{
	int row;

	for (row = 0; row < 6; row++) {
		rowjustify[row] = 255;
		rowpitch[row] = 255;
		rowfix[row] = 255;
		rowcolor[row] = 255;
		rowbackground[row] = 255;
		rainbowrow[row] = 255;
		stripedrow[row] = 255;
		candycanerow[row] = 255;
		rowccolor[row] = 255;
		rowcbg[row] = 255;
	}
}

void updateFontRows()
{
   int row;

    /* if font color is set for any row
       then no special effects */
 	for (row = 0; row < 6; row++) {

		if (rowcolor[row] != 255) {
			if (rainbowrow[row] == 255) rainbowrow[row] = 0;
			if (stripedrow[row] == 255) stripedrow[row] = 0;
			if (candycanerow[row] == 255) candycanerow[row] = 0;
		}
	}
}

/* plot characters in black (color 0) on a white background or
   plot characters in white (color 15) on a black background with either

   - a 7 pixel width (pitch) per character (default) or
   - a 6 pixel pitch

   they can paint them later or use them as-is... since black and white are
   the same between main and auxiliary memory no tranlation is needed */

/* we have room vertically for 6 rows of text */
/* our default horizontal pitch is 7 pixels per character cell for wide characters
   and 6 pixels for narrow characters and spaces.

   12 wide characters = 84 pixels so the right side will clip 4 pixels
   13 narrow characters = 78 pixels so the right and left size will be offset by 1 pixel

   our optional horizontal pitch is 6 pixels per character cell for wide characters
   and 5 pixels for narrow characters and spaces.

   13 wide characters = 78 pixels so the right and left size will be offset by 1 pixel
   16 narrow characters = 80 pixels - no clipping

    default - center justified both horizontally and vertically in window
              starts from the left if string is too long
    optionally - left justified same as when string is too long in default above

    rows = 0-5

    The text file layout is also used to position the text

    1. leaving blanks (padding) to the left will move the first non-space
       character to the right by either 6 or 5 pixels (depending on 7 or 6 pitch).
       Combining this method with FL command option (left justify) allows
       you to manually center some lines and to left justify others.
    2. put empty lines in the file will put 8 lines of space between
       rows of text.
    3. putting less than 6 lines in will cause the "font cluster" display
       to be centered evenly in the vertical axis as well as the horizontal axis


Rationale and Future Thoughts (None Really:)

   This is really a joke in this day and age to be programming a bitmap font
   plotting routine for a display that can only support a maximum of 6 lines
   of 13 or so characters per line.

   Ideally, I would have taken the time to draw a font in the old PC PaintBrush
   format which stored the width and height, or to see if I had a small one around
   here somewhere, but really this isn't worth any further attention.

   A 5 x 6 font would only offer 8 lines of 16 characters and DLGR's mixed mode
   offers 4 text lines of 80 characters per line at the bottom of the screen
   which is a far better alternative for text of any consequence
   so this large font would practically only be used for titling and stuff anyway...

   But in closing, keep in mind that Windows Paint doesn't offer any screaming hell
   of a small font that will convert properly to DLGR, and at least with a good quality
   font like my apple II font it converts properly, it can be painted easily, and
   using this utility, it is simple to do layouts in a text file
   without farting around much.

   */
void lofont(unsigned char *str, int y, int row)
{

  unsigned char bitmsk[]={0x1,0x2,0x4,0x8,0x10,0x20,0x40};

  int x, x1, packet = 40, target, scanline, offset, byte, bit, spaces = 0, width = 80;
  unsigned char bitmap, ch, color = (unsigned char)fontcolor, bgcolor = 255,
           ccolor = (unsigned char) fontccolor, cbg = (unsigned char)fontcbg;

  int pitch = fontpitch, fix = fontfix, justify = fontjustify,
      striped = stripedtext, candycane = candycanetext, rowbow=rainbowtext;

  /* row level over-ride implementation */
  /* these values were initialized at the beginning of the program
     and set from the command line if any have been set, otherwise
     they will be skipped and the "sheet level" values will be used */
  if (rowjustify[row]!=255) justify = rowjustify[row];
  if (rowpitch[row] != 255) pitch = rowpitch[row];
  if (rowfix[row] != 255) fix = rowfix[row];
  if (rowcolor[row] != 255) color = (unsigned char)rowcolor[row];
  if (rowbackground[row] != 255) bgcolor = (unsigned char)rowbackground[row];
  if (rainbowrow[row] != 255) rowbow = rainbowrow[row];
  if (stripedrow[row] != 255) striped = stripedrow[row];
  if (candycanerow[row] != 255) candycane = candycanerow[row];
  if (rowccolor[row] != 255) ccolor = rowccolor[row];
  if (rowcbg[row] != 255) cbg = rowcbg[row];


  if (y > 47) return;

  target = 0;
  for (x = 0;x<17;x++) {
	  if (str[x] == ASCIIZ)break;
	  target++;
  }
  if (target < 1) return;

  if (lores == 1) {
	  if (target > 8) target = 8;
	  width = 40;
	  packet = 20;
  }
  else {
  	  if (target > 16)target = 16; /* see above */
  }

  /* target is only an approximation but pixels will be clipped
     if beyond the right range of the clipping region (pixels 0-79)*/

  /* for our centre-justification calculation we run through
     a simple loop which uses the same logic as the plotting loop and
     decrement our pitch by 1 pixel if we find a narrow character.
  */
  x = 0;
  for (byte=0;byte<target;byte++) {
        ch = str[byte]; x += pitch;
        if (ch < 32)ch = 32;
        if (ch == 32 || ch == 'l' || ch == '1' || ch == '.' || ch == '!') {
			if (x < width) {
				spaces +=fix;
				x-=fix;
			}
		}
  }

  /* center text in the bmp file unless
     too long */
  if (justify == 0) x = 0;
  else if (x > 79 && lores == 0) x = 0;
  else if (x > 39 && lores == 1) x = 0;
  else {
	x1 =  ((target * pitch)- spaces);
	if (x1%2 != 0) x1++;
	if (x1 > width) x1 = width;
	x = packet-(x1/2);
  }


    if (candycane == 1 && striped == 0) {
	   color = ccolor;
	}

	for(scanline=0;scanline<8;scanline++)
	{
		if (striped == 1)color = rainbow[(y/2)%8];
	    if (candycane == 1 && striped == 0 && y%2 == 0) {
			if (color == ccolor) color = cbg;
			else color = ccolor;
		}

		spaces = 0;
		for (byte=0;byte<target;byte++)
		{
			/* calculate the starting pixel for each character in the string */

			x1 = ((byte * pitch) + x)-spaces;
            if (fix == 0 && lores == 0 && pitch == 7 && justify == 1) {
			     /* hard coded centering of boxes */
			     packet = byte * 7;
			     switch(target) {
				   case 11: x1 = 2 + packet; break;
				   case 10: x1 = 6 + packet; break;
				 }
			}
			if (lores == 1)x1 *= 2;
			if (x1 > 82)break;


            /* get the next character in the string */
			ch = str[byte];
			/* my Apple II font modified from Aztec C65 AppleX)
			   starts at ascii character 32 and
			   ends at special character 255

			   It offers the ancient IBM-PC hi-ascii multi-lingual characters
			   as well as the usual low-ascii characters.

			   When I made it (late 80's, early 90's), I narrowed the
			   bitmaps from the PC's 8 X 8 40 character CGA display
			   to fit the Apple II's 7 x 8 40 character HGR display.

			   I provided the special characters in a bi-lingual
			   English-French series of Apple II programs I wrote
			   that were used in Schools across Canada...

			   */

			/* if the value in the ascii string is not in the font
			   blank it... */
			if (ch < 32)ch = 32;

			/* adjust for pitch for each narrow character (within a small
			   subset, mostly punctuation) ...make the most of the small display

			   maybe some day I'll drag the narrow characters to the left
			   of the cell and redo the font to make this work better... I should actually
			   be dragging the mask to the left by one pixel in here too
			   but doing it this way gives the larger characters some breathing room.

			   there's only so much turd-polishing I have the patience for right now.
			   */
			if (ch == 32 || ch == 'l' || ch == '1'|| ch == '.'||ch == '!')spaces +=fix;

			if (rowbow == 1 && candycane == 0 && striped == 0) color = rainbow[byte];

            /* jump across space characters in the string */
			if (ch == 32 && bgcolor == 255)continue;

			offset = (((int)(ch-32)) * 8) + scanline;
			bitmap = __chr255[offset];

            /* jump across blank rows in the font's bitmap */
			if (bitmap == 0)continue;

            /* if double lo-res plot 1 pixel */
            /* if lo-res, plot 2 pixels */
			for (bit=0;bit<7;bit++)
			{
				if (lores == 1) {
					if (bitmap & bitmsk[bit]) {
						setpixel(x1-2,y,color);
					}
					else {
						if (bgcolor != 255) setpixel(x1-2,y,bgcolor);
					}
					x1++;
					if (x1 > 82)break;
					if (bitmap & bitmsk[bit]) {
						setpixel(x1-2,y,(unsigned char)color);
					}
					else {
						if (bgcolor != 255) setpixel(x1-2,y,bgcolor);
					}
					x1++;
					if (x1 > 82)break;
				}
				else {
					if (bitmap & bitmsk[bit]) {
						setpixel(x1-1,y,color);
					}
					else {
						if (bgcolor != 255) setpixel(x1-1,y,bgcolor);
					}
					x1++;
					if (x1 > 80)break;
				}

			}

		}
		y++;
		if (y > 47)break;
	}

}


void DrawBorder()
{
	int x, y;

	if (border == 255) return;

	for (x = 0; x < 80; x++) {
	    setpixel(x,0,(unsigned char)border);
	    setpixel(x,47,(unsigned char)border);
	}

	for (y = 1; y < 47; y++) {
	    setpixel(0,y,(unsigned char)border);
	    setpixel(79,y,(unsigned char)border);
	    if (lores == 1) {
			setpixel(1,y,(unsigned char)border);
			setpixel(78,y,(unsigned char)border);
		}
	}
}

/* this tab filter snippet is from my command-line
   commando filter program "untab" ...modified to suit this occasion */
int untab(unsigned char *src, unsigned char *dest, int tabstop, int bufmax)
{
    unsigned char c;
    int i, idx, jdx, spaces;

    // if a valid numeric argument use the tabstop value
    // otherwise default to 8 space max per tab
    if (tabstop < 2 || tabstop > 8) tabstop = 8;

    spaces = tabstop;   // initialize spaces to a full tab

	for(idx=0,jdx=0;;idx++) {

		c = src[idx];
		if(c==(unsigned char)10 || c==(unsigned char)13) c = ASCIIZ;
		if (c == ASCIIZ) {
			dest[jdx] = c;
			break;
		}

		 if(c==(unsigned char)9) // tab found, so print remaining spaces
		  {
			for(i=0;i<spaces;i++) {
				dest[jdx] = ' ';
				jdx++;
				if (jdx < bufmax) continue;
				dest[jdx] = ASCIIZ;
				break;
			}
			if (jdx < bufmax) {
				spaces = tabstop;
				continue;
			}
			break;
		  }

           spaces--;   // other character found, one less space required
                       // until next tabstop
           if(spaces == 0)spaces = tabstop; // spaces are all used so
                                            // reset to a full tabstop
           dest[jdx] = c; jdx++;
           if (jdx < bufmax) continue;
           dest[jdx] = ASCIIZ;
           break;
	}

	return jdx;
}


int TEXT_Read(unsigned char *basename)
{

	FILE *fp;
	unsigned char infile[128], src[130], dest[130];
	int cnt, y, i;

	sprintf(infile,"%s.TXT",basename);
	if (FileExists(infile) == 0) {
		sprintf(infile,"%s.ASC",basename);
		if (FileExists(infile) == 0) return INVALID;
	}

	fp = fopen(infile,"r");
	if (NULL == fp) return INVALID;

	/* count the lines in the file, max 6 lines */
	cnt = 0;
	for (i=0; i< 6; i++) {
	  if (fgets((unsigned char *)&src[0],128,fp) == NULL) break;
	  cnt++;
	}

	/* return if a fail */
	if (cnt == 0) {
		fclose(fp);
		return INVALID;
	}

	/* centre text vertically in the bmp based on our line
	   count... we need to do that here which the other part of
	   the reason for doing the line count besides checking for
	   an empty file. the font routine centres the text horizontally
	   so no worries about that... */
	y = (48 - (cnt * 8))/2;
	rewind(fp);

    /* if overlaying an image with text the background is
       not set and the text plots on-top of the image already
       in the bmp buffer - otherwise set the background color */
    if (textoverlay == 0) {
		memset(&bmpbuf[0],(char)fontbackground,3840); /* clear write buffer */
	    DrawBorder(); /* if not overlaying we draw a border on the background
	                     if the border color has been set */
	}
	for (i=0; i< 6; i++) {
	  if (fgets((unsigned char *)&src[0],128,fp) == NULL) break;
	  src[16] = ASCIIZ;
	  /* just untab 16 characters... */
	  if (untab((unsigned char *)&src[0],(unsigned char *)&dest[0], 8, 16) > 0) {
	  		dest[16] = ASCIIZ;
	  		lofont((unsigned char *)&dest[0],y,i);
	  }
	  y+= 8;
	}
	fclose(fp);

return SUCCESS;
}

void main(int argc, char **argv)
{

  int status, idx, jdx, bmp = 0, dlo = 0, txt = 0,temp,row;
  unsigned char fname[128],sname[128],version[128], buf[128], b1[2], c,d,e,licence=1;

  system("cls");
  puts(szTextTitle);
  InitFontRows(); /* initilialize text file defaults */

  if(argc == 1) {
    puts("Command line Usage is:");
    puts("  \"AWINDLO AppleWinScreenGrab.BMP\" or");
    puts("  \"AWINDLO MyNativeAppleII.DLO\" (or MyNativeAppleII.DL1)");
    puts("  if DL1, both files in DL1,DL2 file pair must be present...");
    puts("  \"AWINDLO My6Line.TXT\"   (or My6Line.ASC)");
    puts("  \"AWINDLO My6Line.TXT F6\" (narrower font option)");
    puts("  \"AWINDLO My6Line.TXT FL\" (left justify font option)");
    puts("  \"AWINDLO My6Line.TXT FC[color 0-16]\" (font color)");
    puts("Output default: 160 x 128 24-bit bmp with color palette");
    puts("           and: DLO and DL1,DL2 Apple II Native DLGR files.");
    puts("Output Palette Option:(turns palette off - default is on)");
    puts("  \"AWINDLO AppleWinScreenGrab.BMP P\"");
    puts("Output Size Option:(turns double scale off - default is on)");
    puts("  \"AWINDLO AppleWinScreenGrab.BMP S\"");
    puts("AppleWin Version Option: (force version - default is autodetect)");
    puts("	version 120 color mapping \"AWINDLO AppleWinScreenGrab.BMP 120\"");
    puts("	version 117 color mapping \"AWINDLO AppleWinScreenGrab.BMP 117\"");
    puts("	version 116 color mapping \"AWINDLO AppleWinScreenGrab.BMP 116\"");
    puts("Example: \"AWINDLO AppleWinScreenGrab.BMP P S 120\"");
    puts("  output:  tiny 80 x 48 bmp without palette in AWin 1.20 colors.");
    puts("FileName must use 8.3 character MS-DOS Naming... no Long Names!");
    idx = 0;
    do {
		if (idx > 0) printf("%s could not be opened!\n",fname);
    	printf("Enter Input FileName (Blank to Exit): ");
    	gets(fname);
    	if (fname[0] == ASCIIZ) {
	  		puts("Have a Nice Dos!");
      		exit(1);
  		}
  		idx++;
	} while (FileExists(fname) == 0);
    printf("Enter Version Number (Blank for None) : ");
    gets(version);
  }
  else {
    strcpy(fname, argv[1]);
    /* check if input file exists, exit if not */
    if (FileExists(fname) == 0) {
		/* the following is not accurate because this program accepts pathed filnames
		   but warn them anyway... */
		printf("%s could not be opened!\n",fname);
		puts("Here are some things to look for besides spelling errors...");
		if (strlen(fname) > 12)
		  puts("FileName must use 8.3 character MS-DOS Naming...no Long Names!");
		puts("File must exist, must not be locked,\n"
			 "and you must have full access including read and write privileges.\nExiting");
    	exit(1);

	}
	version[0] = ASCIIZ;
 	for (idx = 2; idx < argc; idx++) {
		// switch char
		if (argv[idx][0] == '-') jdx = 1;
		else jdx = 0;
		strcpy(buf, (char *)&argv[idx][jdx]);
		if ((buf[0] == '1' || buf[0] == '2') && buf[1] == ASCIIZ) {
			if (buf[0] == '1') lores = 1;
		}
		else if (buf[0] > 47 && buf[0] < 58) strcpy(version, buf);
		else {
		   c = toupper(buf[0]);
		   if (c == 'B') {
			   if (buf[1] > 47 && buf[1] < 58) {
				   temp = atoi((char *)&buf[1]);
				   if (temp > -1 && temp < 16) border = temp;
			   }
		   }
		   if (c == 'T') textoverlay = 1;
		   if (c == 'P') outputpalette = 0;
		   if (c == 'S') outputscale = 1;
		   if (c == 'Q') silent = 1;
		   if (c == 'L') licence = 0;
		   if (c == 'F') {
			   c = toupper(buf[1]);
			   if (c == 'L')fontjustify = 0; /* left justify - x = 0 */
			   if (c == 'M')fontjustify = 1; /* middle - default */
			   if (c == '5')fontpitch = 5;
			   if (c == '6')fontpitch = 6;
			   if (c == '7')fontpitch = 7;
			   if (c == 'F')fontfix = 0;
			   if (c == 'C') {
				   if (buf[2] > 47 && buf[2] < 58) {
				   		temp = atoi((char *)&buf[2]);
				   		if (temp > -1 && temp < 16) fontcolor = temp;
				   		if (temp == 16)rainbowtext = 1;
				   		if (temp == 17)stripedtext = 1;
				   		if (temp == 18)candycanetext = 1;
				   }
				   if (toupper(buf[2]) == 'E') {
					   if (buf[3] > 47 && buf[3] < 58) {
							temp = atoi((char *)&buf[3]);
							if (temp > -1 && temp < 16) fontccolor = temp;
					   }
				   }
			   }
			   if (c == 'B') {
				   if (buf[2] > 47 && buf[2] < 58) {
				   		temp = atoi((char *)&buf[2]);
				   		if (temp > -1 && temp < 16) fontbackground = temp;
				   }
				   if (toupper(buf[2]) == 'E') {
					   if (buf[3] > 47 && buf[3] < 58) {
							temp = atoi((char *)&buf[3]);
							if (temp > -1 && temp < 16) fontcbg = temp;
					   }
				   }
			   }
		   }
		   if (c == 'R' && (buf[1] > 47 && buf[1] < 54)) {

			   b1[1] = ASCIIZ;
			   b1[0] = buf[1];

			   row = atoi((char *)&b1[0]);
			   c = toupper(buf[2]);

			   if (c == 'L')rowjustify[row] = 0; /* left justify - x = 0 */
			   if (c == 'M')rowjustify[row] = 1; /* middle - default */
			   if (c == '5')rowpitch[row] = 5;
			   if (c == '6')rowpitch[row] = 6;
			   if (c == '7')rowpitch[row] = 7;
			   if (c == 'F')rowfix[row] = 0;
			   if (c == 'U')rowfix[row] = 1;
			   if (c == 'C') {
				   if (buf[3] > 47 && buf[3] < 58) {
						temp = atoi((char *)&buf[3]);
						if (temp > -1 && temp < 16) rowcolor[row] = temp;
						if (temp == 16)rainbowrow[row] = 1;
						if (temp == 17)stripedrow[row] = 1;
						if (temp == 18)candycanerow[row] = 1;
				   }
				   if (toupper(buf[3]) == 'E') {
					   if (buf[4] > 47 && buf[4] < 58) {
							temp = atoi((char *)&buf[4]);
							if (temp > -1 && temp < 16) rowccolor[row] = temp;
					   }
				   }

			   }
			   if (c == 'B') {
				   if (buf[3] > 47 && buf[3] < 58) {
						temp = atoi((char *)&buf[3]);
						if (temp > -1 && temp < 16) rowbackground[row] = temp;
				   }
				   if (toupper(buf[3]) == 'E') {
					   if (buf[4] > 47 && buf[4] < 58) {
							temp = atoi((char *)&buf[4]);
							if (temp > -1 && temp < 16) rowcbg[row] = temp;
					   }
				   }

			   }
		   }
	   }
	}
  }

  applewinversion = atoi(version);
  updateFontRows();

  /* get basename and check for valid filetypes */
  jdx = 999;
  for (idx = 0; fname[idx]!=ASCIIZ; idx++) {
  	 if (fname[idx] == '.') jdx = idx;
  }

  if (jdx != 999) {
      strcpy(sname, fname);
	  sname[jdx] = ASCIIZ; jdx++;
	  c = toupper(fname[jdx]); jdx++;
	  d = toupper(fname[jdx]); jdx++;
	  e = toupper(fname[jdx]);

      jdx = 999;
	  if (c == 'B' && d == 'M' && e == 'P') bmp = jdx = 1;
	  if (c == '2' && d == '5' && e == '6') bmp256 = bmp = jdx = 1;
	  if (c == 'D' && d == 'L' && e == 'O') dlo = jdx = 1;
	  if (c == 'D' && d == 'T' && e == 'O') dlo = jdx = 1;
	  if (c == 'D' && d == 'L' && e == '1') dlo = jdx = 1;
	  if (c == 'D' && d == 'L' && e == '2') dlo = jdx = 1;
	  if (c == 'S' && d == 'L' && e == 'O') dlo = jdx = 1;
	  if (c == 'S' && d == 'T' && e == 'O') dlo = jdx = 1;
	  if (c == 'S' && d == 'L' && e == '2') dlo = jdx = 1;
	  if (c == 'T' && d == 'X' && e == 'T') txt = jdx = 1;
	  if (c == 'A' && d == 'S' && e == 'C') txt = jdx = 1;
  }

  if (jdx == 999) {
     printf("%s is an Unsupported FileType.\nSupported Extensions are BMP,DLO,DL1,DL2,TXT or ASC.\n", fname);
     exit(1);

  }
   if (licence)
   puts("Licence Agreement\n"
		"-----------------\n"
		"You have a royalty-free right to use, modify, reproduce and\n"
		"distribute this program, its source code, and any related stuff of mine\n"
		"it came with in any way you find useful, provided that you agree that I,\n"
		"Bill Buckels, have no warranty obligations or liability resulting from said\n"
		"use and distribution in any way whatsoever. If you don't agree, remove this\n"
		"program, and all the stuff it came with from your computer immediately.\n"
		"(To suppress this message use the L (Licence) Command Line Option.)");



  if (txt == 1) {
	  status = TEXT_Read(sname);
	  if (status == SUCCESS) {
			/* save bmp file and native apple II files */
			status = savelofragment(sname);
			if (lores == 0) {
				if (status == SUCCESS)printf("%s.DL1 & .DL2 and %s.DLO Saved!\n",sname,sname);
				else printf("Error saving %s.DL1 & .DL2 and %s.DLO!\n",sname,sname);
			}
			else {
				if (status == SUCCESS)printf("%s.SL2 and %s.SLO Saved!\n",sname,sname);
				else printf("Error saving %s.SL2 and %s.SLO!\n",sname,sname);
			}
			/* see notes below */
			BackUpOldFile(sname, "BMP", "DIB");
			status = save_to_bmp24(sname);
			if (status == SUCCESS) printf("%s.BMP Created!\n",sname);
			else printf("Error creating %s.BMP!\n",sname);
  	  }
  	  else
  	  	printf("%s is an Unsupported Format or cannot be opened.\n", fname);
  }

  if (dlo == 1) {
  		status = DLO_Read(sname,fname);

		if (status == SUCCESS) {
			DrawBorder();
			if (textoverlay == 1)TEXT_Read(sname);
			/* save bmp file only... do not overwrite native apple II input files ...
			   and do not overwrite bmp files of the same basename if they exist.

			   we are using two naming conventions for "backups" of
			   bmps in this program: .256 and .DIB.

			   the first naming convention is to rename an applewin bmp with a .256
			   extension when we process an applewin bmp. since we don't accept our own
			   bmp's for processing (that is done by BMP2LO), a .256 will likely be created
			   only once unless the user uses the same name again.

			   the second naming convention which is done with input formats like text and
			   apple II files, is to rename our own bmp with a .DIB extension.
			   we are reasonably safe to rename our own bmp's if any have a basename that
			   conflicts with a native apple II file and they are still in the working directory.

			   since we can't fill-up the user's computer with back-up after back-up
			   I decided just to warn them each time we rename or remove a file.

			   Windows Paint will directly open a .DIB, but file assocations need to
			   be set-up for a .256. Regardless I consider what I have done here reasonably safe
			   especially for a programmer's utility.

			   However potential for conflict still exists if the user is messy and creates
			   same-named DLO's, etc. outside of here and doesn't clean-up their AwinDLO
			   working directory between sessions.

			   I didn't really want to put a bunch of warnings in place, and most of
			   my utilities just overwrite without asking (including BMP2LO) but in this case
			   we are doing something that is special enough, in the end I decided to give
			   them warnings and one last chance to clean-up old back-ups.

			   Beyond that, they are on their own, as in life!

			   */
			BackUpOldFile(sname, "BMP", "DIB");
			status = save_to_bmp24(sname);
			if (status == SUCCESS) printf("%s.BMP Created!\n",sname);
			else printf("Error creating %s.BMP!\n",sname);

		}
		else
			printf("%s is an Unsupported Format or cannot be opened.\n", fname);

  }

  if (bmp == 1) {
		status = AppleWinRead(sname);

		if (status == SUCCESS) {
			DrawBorder();
			if (textoverlay == 1)TEXT_Read(sname);
			/* save bmp file and native apple II files */
			status = savelofragment(sname);
			if (lores == 0) {
				if (status == SUCCESS)printf("%s.DL1 & .DL2 and %s.DLO Saved!\n",sname,sname);
				else printf("Error saving %s.DL1 & .DL2 and %s.DLO!\n",sname,sname);
			}
			else {
				if (status == SUCCESS)printf("%s.SL2 and %s.SLO Saved!\n",sname,sname);
				else printf("Error saving %s.SL2 and %s.SLO!\n",sname,sname);
			}

			status = save_to_bmp24(sname);
			if (status == SUCCESS)printf("%s.BMP Converted/Saved!\n",sname);
			else printf("Error converting/saving %s.BMP!",sname);
		}
		else
			printf("%s is an Unsupported Format or cannot be opened.\n", fname);
  }


  if (backup && silent == 0)
     puts("WARNING! \"Back-Ups\" of duplicate filenames have been deleted!\n"
          "These deleted files are probably unrecovereable, so I hope\n"
          "you either don't care or have another copy. If that is the case\n"
          "I understand why you have ignored AWINDLO's Back-Up messages in the past,\n"
          "and I apologize for this reminder... Otherwise you seem to be a little\n"
          "behind in your housekeeping. While it is probably too late to save the file(s)\n"
          "you have already lost... and keeping in mind that you are using AWINDLO\n"
          "entirely at your own risk (like any other user of this program)\n"
          "*IF* you want to save away (or review) your other AWINDLO \"Back-Ups\"...\n"
          "*NOW* (or even sooner) might be a very good time! (Later will be too late!)\n"
          "\"Back-Ups\" of the same basename *WILL* always be overwitten!\n"
          "Use the Q (quiet) command line switch to avoid *ALL* back-up messages.\n"
          "And may HAL be merciful!");

  puts("Have a Nice Dos!");
  exit(0);

}
