/* ---------------------------------------------------------------------
HMONSTER (HMON.C) (C) Copyright Bill Buckels 2013.
All Rights Reserved.

Licence Agreement
-----------------

You have a royalty-free right to use, modify, reproduce and
distribute this source code in any way you find useful, provided
that you agree that Bill Buckels has no warranty obligations or
liability resulting from said distribution in any way whatsoever. If
you don't agree, remove this source code and related files from your
computer now.

Written by: Bill Buckels
Email:      bbuckels@escape.ca

Purpose 1:  "Round Trip" utility to convert between Windows
			Monochrome and Color BMP files and Apple II
			Monochrome and Color HGR files.

Purpose 2:  "One-Way" utility to convert 320 x 200 x 4-Color GWBASIC
			BSaved Images (now "extinct")to Apple II HGR files.

Synopsis:


Apple II Output

This utility converts from the following IBM-PC (MS-DOS and Windows) Graphics
file formats to Apple II HGR files:

2 Color (Monochrome) BMP Files in 280 x 192 or 320 x 200 resolution
16 Color (VGA) BMP Files in 280 x 192 or 320 x 200 resolution
256 Color BMP Files in 280 x 192 or 560 x 384 resolution
24 bit BMP Files in 280 x 192 or 560 x 384 resolution
320 x 200 x 4-Color GWBASIC BSaved Images (now "extinct")

The Apple II HGR files it creates are limited to 280 x 192
("fullscreen" Graphics-Only) and 280 x 160 (Mixed Text and Graphics size)
Images in the following formats:

BIN - 280 x 192 Color Images in BSaved (BIN) format
BOT - 280 x 192 Color Images in Raster (RAG) format
TOP - 280 x 160 Color Images in Raster (RAG) format

BIN - 280 x 192 Monochrome Images in BSaved (BIN) format
BOT - 280 x 192 Monochrome Images in Raster (RAG) format
TOP - 280 x 160 Monochrome Images in Raster (RAG) format

Several Commandline Options are available.

BMP Output

This Utility Converts from the following Apple II Graphics file formats
to Windows 280 x 192 x 16 Color BMP files:

BIN - 280 x 192 Color Images in BSaved (BIN) format
BOT - 280 x 192 Color Images in Raster (RAG) format
TOP - 280 x 160 Color Images in Raster (RAG) format

This Utility Converts from the following Apple II Graphics file formats
to Windows 280 x 192 x 2 Color (Monochrome) BMP files:

BIN - 280 x 192 Monochrome Images in BSaved (BIN) format
BOT - 280 x 192 Monochrome Images in Raster (RAG) format
TOP - 280 x 160 Monochrome Images in Raster (RAG) format

Please read the documentation, the source code, and run the program
for more information.

Bill Buckels
March 12,2013

Revision     : 1.0 First Release
-----------------------------------------------------------------------
Written in Large Model 16 bit Microsoft C (MSC) Version 8.00c
Note: Run in an MS-DOS emulator like DOSBox if you can't run it raw.
----------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

/* Bitmap Header structures */
typedef struct tagRGBQUAD
{
    unsigned char    rgbBlue;
    unsigned char    rgbGreen;
    unsigned char    rgbRed;
    unsigned char    rgbReserved;
} RGBQUAD;

typedef struct tagBITMAPINFOHEADER
{
    unsigned long   biSize;
    unsigned long   biWidth;
    unsigned long   biHeight;
    unsigned int    biPlanes;
    unsigned int    biBitCount;
    unsigned long   biCompression;
    unsigned long   biSizeImage;
    unsigned long   biXPelsPerMeter;
    unsigned long   biYPelsPerMeter;
    unsigned long   biClrUsed;
    unsigned long   biClrImportant;
} BITMAPINFOHEADER;

/* constants for the biCompression field */
#define BI_RGB      0L
#define BI_RLE8     1L
#define BI_RLE4     2L

typedef struct tagMONOBITMAPINFO
{
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD          bmiColors[2];
} MONOBITMAPINFO;

typedef struct tagVGABITMAPINFO
{
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD          bmiColors[16];
} VGABITMAPINFO;


typedef struct tagBITMAPFILEHEADER
{
    unsigned char   bfType[2];
    unsigned long   bfSize;
    unsigned int    bfReserved1;
    unsigned int    bfReserved2;
    unsigned long   bfOffBits;
} BITMAPFILEHEADER;


/* translation table for remapping of bmp colors */
unsigned char bmpmapinfo[16] = {255,255,255,255,255,255,255,255,
                                255,255,255,255,255,255,255,255};


/* 16 Color BMP style palette (Windows XP) */
/* notice that these johnny-come-latelies reversed GRAY and WHITE */
enum {  VGA_BLACK = 0,
        VGA_BLUE,
        VGA_GREEN,
        VGA_CYAN,
        VGA_RED,
        VGA_MAGENTA,
        VGA_BROWN,
        VGA_GRAY,
        VGA_WHITE,
        VGA_LBLUE,
        VGA_LGREEN,
        VGA_LCYAN,
        VGA_LRED,
        VGA_LMAGENTA,
        VGA_YELLOW,
        VGA_BWHITE,
        NUM_VGA_COLORS};

 enum { CGA_BLACK = 0,
        CGA_CYAN,
        CGA_MAGENTA,
        CGA_WHITE,
        NUM_CGA_COLORS};


/* these constants are what I use in this program but have a limited
   meaning in reality and therefore should not be taken as anything but
   process control ordinals... they do not hold the mystery of the
   universe and are simply an ordering of the 6 Apple II HGR colors */

enum hgrcolors{
	HBLACK=0,
	HGREEN,
	HVIOLET,
	HORANGE,
	HBLUE,
	HWHITE};

/* static structures for processing input file */
BITMAPFILEHEADER BitMapFileHeader;
RGBQUAD          sbmp[256]; /* super vga - applewin */
VGABITMAPINFO    bmp;
MONOBITMAPINFO   mbmp;
BITMAPINFOHEADER bmi;


#define ASCIIZ 0


unsigned char *szTextTitle =
    "HMONSTER(C) Version 1.0 Copyright Bill Buckels 2013\n"
    "All Rights Reserved.";

#define SUCCESS  0
#define VALID    SUCCESS
#define FAILURE  -1
#define INVALID  FAILURE

/* a GWBASIC compatible bsaved image format descriptor */
unsigned char BSAVED_header[7]={
    '\xfd','\x00','\xb8','\x00','\x00','\x00','\x40'};

unsigned char msk[]={0x80,0x40,0x20,0x10,0x8,0x4,0x2,0x1};

/* provides base address for page1 hires scanlines  */
unsigned HB[]={
0x2000, 0x2400, 0x2800, 0x2C00, 0x3000, 0x3400, 0x3800, 0x3C00,
0x2080, 0x2480, 0x2880, 0x2C80, 0x3080, 0x3480, 0x3880, 0x3C80,
0x2100, 0x2500, 0x2900, 0x2D00, 0x3100, 0x3500, 0x3900, 0x3D00,
0x2180, 0x2580, 0x2980, 0x2D80, 0x3180, 0x3580, 0x3980, 0x3D80,
0x2200, 0x2600, 0x2A00, 0x2E00, 0x3200, 0x3600, 0x3A00, 0x3E00,
0x2280, 0x2680, 0x2A80, 0x2E80, 0x3280, 0x3680, 0x3A80, 0x3E80,
0x2300, 0x2700, 0x2B00, 0x2F00, 0x3300, 0x3700, 0x3B00, 0x3F00,
0x2380, 0x2780, 0x2B80, 0x2F80, 0x3380, 0x3780, 0x3B80, 0x3F80,
0x2028, 0x2428, 0x2828, 0x2C28, 0x3028, 0x3428, 0x3828, 0x3C28,
0x20A8, 0x24A8, 0x28A8, 0x2CA8, 0x30A8, 0x34A8, 0x38A8, 0x3CA8,
0x2128, 0x2528, 0x2928, 0x2D28, 0x3128, 0x3528, 0x3928, 0x3D28,
0x21A8, 0x25A8, 0x29A8, 0x2DA8, 0x31A8, 0x35A8, 0x39A8, 0x3DA8,
0x2228, 0x2628, 0x2A28, 0x2E28, 0x3228, 0x3628, 0x3A28, 0x3E28,
0x22A8, 0x26A8, 0x2AA8, 0x2EA8, 0x32A8, 0x36A8, 0x3AA8, 0x3EA8,
0x2328, 0x2728, 0x2B28, 0x2F28, 0x3328, 0x3728, 0x3B28, 0x3F28,
0x23A8, 0x27A8, 0x2BA8, 0x2FA8, 0x33A8, 0x37A8, 0x3BA8, 0x3FA8,
0x2050, 0x2450, 0x2850, 0x2C50, 0x3050, 0x3450, 0x3850, 0x3C50,
0x20D0, 0x24D0, 0x28D0, 0x2CD0, 0x30D0, 0x34D0, 0x38D0, 0x3CD0,
0x2150, 0x2550, 0x2950, 0x2D50, 0x3150, 0x3550, 0x3950, 0x3D50,
0x21D0, 0x25D0, 0x29D0, 0x2DD0, 0x31D0, 0x35D0, 0x39D0, 0x3DD0,
0x2250, 0x2650, 0x2A50, 0x2E50, 0x3250, 0x3650, 0x3A50, 0x3E50,
0x22D0, 0x26D0, 0x2AD0, 0x2ED0, 0x32D0, 0x36D0, 0x3AD0, 0x3ED0,
0x2350, 0x2750, 0x2B50, 0x2F50, 0x3350, 0x3750, 0x3B50, 0x3F50,
0x23D0, 0x27D0, 0x2BD0, 0x2FD0, 0x33D0, 0x37D0, 0x3BD0, 0x3FD0};

/* apple II HGR screen buffer */
unsigned char hgrbuf[8192];

/* work buffers...
   only the first 35 bytes are used to process monochrome bmp
   scanlines but for 16 color bmp's a 160 byte read buffer
   and 320 byte process buffer are required.
*/
unsigned char svgabuf[320];
unsigned char bmpscanline[320];
unsigned char buf280[280];
unsigned char pal280[280];

/* commandline options */
unsigned char xPalette = 0, a2palette = 0x80, a2reverse = 0, a2shift = 0, monoshift = 0,
              a2center = 0, a2switch = 0, a2top = 0, a216 = 0, a256 = 0;

unsigned externalmap = 0, applewindex =0, applewinversion = 0;


/* when a color bmp is output to an Apple II HGR file and the xPalette Option
   is used, palettebits sets the hibits of each byte in the hgr output
   raster based on color dominance by doing a pixel count of green-violet
   vs. orange-blue for every two bytes or every byte in the hgr raster.
   orange and blue pixels take precedence over green and purple pixels */
unsigned char palettebits[40];
unsigned char palettebuf[320];

void clearpalettebits()
{
   int x;

   for (x = 0; x < 40; x++) palettebits[x] = a2palette;
}

/* decodes bmp scanline from file */
int ibmbits()
{
     int i,j,k;

     j=0;
     for(i=0;i<35;i++)
     {

        for(k=0;k<8;k++)
        {
            if(bmpscanline[i]&msk[k])buf280[j] = 1;
            else buf280[j] = 0;
            j++;
        }
     }
     return SUCCESS;

}

/* decodes bmp shifted scanline from file
   monochrome source only */
int ibmshiftbits()
{
     int i,j,k,p, count;
     unsigned char pixel;

     /* for monochrome bmps that are wider than 280 (i.e. 320)
        we shift into the bitmask itself and handle single pixels
        and offsetting for IBM color translation control on the
        a color HGR display without upsetting the monochrome
        representation */
     for(i=0,j=0,p=0,count=0;i<37;i++)
     {

        for(k=0;k<8;k++)
        {

            if(bmpscanline[i]&msk[k])pixel = 1;
            else pixel = 0;
            j++;


            if (count < monoshift) {
				count++;
				continue;
			}

            if (p < 280)buf280[p] = pixel;
            p++;
        }
     }
     return SUCCESS;

}



/* encodes apple II hgr scanline into buffer */
/* applies palette bit and reverse video options */
int applebits(int y)
{
	    int xoff,idx,jdx;
	    unsigned char *ptr, bits[7], x1, palettebit;

	    xoff = HB[y]-0x2000;
	    ptr  = (unsigned char *) &hgrbuf[xoff];

        xoff = 0;
        for (idx = 0; idx < 40; idx++) {

            for (jdx = 0; jdx < 7; jdx++) {
				bits[jdx] = buf280[xoff]; xoff++;
			}
			palettebit = palettebits[idx];
			/* palette bit options
			   global option - use green violet instead of orange blue default

			   color bmp option - xPalette option (P14 or P7)
			   this option if used is added to the global option when a color bmp
			   is output to an Apple II HGR file...

			   2 - palette bits will be set for every 14 bits or
			   1 - palette bit will be set for every 7 bits

			   this option preserves pixel colors somewhat


			   */
			x1 = (palettebit | bits[6]<<6|bits[5]<<5|bits[4]<<4|
                        bits[3]<<3|bits[2]<<2|bits[1]<<1|bits[0]);

			/* reverse video option */
            if (a2reverse == 1)  ptr[idx] = x1 ^ 0x7f;
            else ptr[idx] = x1;
		}
		return SUCCESS;

}


/* output is raw only... use raxpack.exe to compress
   file if desired... too much duplicate code otherwise
   and debatable whether anyone will want compressed
   monochrome images much or if a busy image will
   compress much anyway...

   however if converting monochrome forms of
   IBM color images like IBM 320 x 200 BSAVED files
   you may actually want to try raxpack. */
int makebot(FILE *fp, FILE *fp2)
{
    int x, y, xoff;
    unsigned char *ptr;

    for (y = 0; y < 192; y++) {
	    xoff = HB[y] - 0x2000;
    	ptr  = (unsigned char *) &hgrbuf[xoff];
		fwrite(ptr,1,40,fp);
		if (a2top == 0) continue;
		if (y < 160) fwrite(ptr,1,40,fp2);
	}
    return SUCCESS;
}


/* save apple II hgr files... */
int savehgrfiles(unsigned char *basename)
{
	FILE *fp, *fp2=NULL;
	unsigned char binfile[128], ragfile[128], topfile[128];

	sprintf(binfile,"%s.BIN",basename);
	fp = fopen(binfile,"wb");
	if (NULL == fp)return INVALID;
	fwrite(hgrbuf,1,8192,fp);
	fclose(fp);

    sprintf(ragfile,"%s.BOT",basename);
	fp = fopen(ragfile,"wb");
	if (NULL == fp) {
		remove(binfile);
		return INVALID;
	}

    if (a2top == 1) {
    	sprintf(topfile,"%s.TOP",basename);
 		fp2 = fopen(topfile,"wb");
		if (NULL == fp2) {
			fclose(fp);
			remove(ragfile);
			remove(binfile);
			return INVALID;
		}
	}

    /* width in bytes x height in rasters */
    fputc(40,fp);
    fputc(192,fp);
    if (a2top == 1) {
    	fputc(40,fp2);
    	fputc(160,fp2);
	}
	makebot(fp,fp2);
	fclose(fp);

	if (a2top == 1)fclose(fp2);

	return SUCCESS;
}


/* 54 byte headers followed by colormaps of
   4 bytes per color */

/* note: the length of the 256 color map is
         1024 (256 * 4 bytes per color)

         header = 54 + 1024 = 1078

*/

/* header for monochrome BMP file output */

/* 280 x 192 - the default output format */
unsigned char mono192[62] ={
0x42, 0x4D, 0x3E, 0x1B, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x3E, 0x00, 0x00, 0x00, 0x28, 0x00,
0x00, 0x00, 0x18, 0x01, 0x00, 0x00, 0xC0, 0x00,
0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x1B, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00};

/* header for 16 color BMP file output */

/* 280 x 192 - the optional output format */
unsigned char color280[118] ={
0x42, 0x4D, 0x76, 0x69, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x76, 0x00, 0x00, 0x00, 0x28, 0x00,
0x00, 0x00, 0x18, 0x01, 0x00, 0x00, 0xC0, 0x00,
0x00, 0x00, 0x01, 0x00, 0x04, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x69, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x80,
0x00, 0x00, 0x00, 0x80, 0x80, 0x00, 0x80, 0x00,
0x00, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x80,
0x00, 0x00, 0x80, 0x80, 0x80, 0x00, 0xC0, 0xC0,
0xC0, 0x00, 0x00, 0x00, 0xFF, 0x00, 0x00, 0xFF,
0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0xFF, 0x00,
0x00, 0x00, 0xFF, 0x00, 0xFF, 0x00, 0xFF, 0xFF,
0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00};


/* read bmp file in supported resolutions */
int BMPMono_Read(unsigned char *basename, int x, int y)
{

	unsigned int temp, temp2;
	unsigned char bmpfile[128], buf[40], *ptr;
	int i, j, target, packet = INVALID;
	FILE *fp;

    sprintf(bmpfile,"%s.BMP",basename);
	fp=fopen(bmpfile,"rb");

	if (NULL == fp) return INVALID;

    /* check the bmp header */
    /* read the header stuff into the appropriate structures */
    fread((char *)&BitMapFileHeader.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&mbmp.bmiHeader.biSize,
                 sizeof(MONOBITMAPINFO),1,fp);

    if (mbmp.bmiHeader.biCompression==BI_RGB &&
        BitMapFileHeader.bfType[0] == 'B' && BitMapFileHeader.bfType[1] == 'M' &&
        mbmp.bmiHeader.biPlanes==1  && mbmp.bmiHeader.biBitCount==1) {

        if (mbmp.bmiHeader.biWidth == 280 && mbmp.bmiHeader.biHeight == 192) {
			x = 0;
			y = 0;
			packet = 36;
		}
        if (mbmp.bmiHeader.biWidth == 320 && mbmp.bmiHeader.biHeight == 200) {
			packet = 40;
			if (y < 0) y = 0;
			if (y > 7) y = 8;
			y = 8 - y;
			if (x < 1) x = 0;
			if (a2shift != 0) x++;
			/* adjust right edge if out of bounds */
			while (x > 27) x-=2; /* preserve IBM 2 bit patterns */
			/* save the starting bit in the first byte */
			monoshift = x%8;
			x = x/8;
		}
	}

    if (packet == INVALID) { fclose(fp); return INVALID;}

	memset(&hgrbuf[0],0,8192); /* clear write buffer */
	clearpalettebits();

    if (packet == 40) {
    	for (i=0; i < y; i++) fread((char *)&buf[0],1,packet,fp);
	}


    ptr = (unsigned char *)&buf[x];

    if (monoshift == 0) target = 36;
    else target = 37;

    /* remap the image to apple2 hires mono format*/
	for (i = 192; i > 0; i--) {
		  fread((char *)&buf[0],1,packet,fp);
	  	  for (j = 0; j < target; j++) {
			  bmpscanline[j] = ptr[j];
		  }
		  if (monoshift == 0)ibmbits(); // decode 8 bit
		  else ibmshiftbits();

		  applebits(i-1); // encode 7 bit
	}

	fclose(fp);

  	return SUCCESS;
}


/* I have not gone to the time and effort to
   match the default palettes for 256 color and 24 bit bmp files
   created in Windows Paint.

   I do provide support for my own clipshop utility's default colors,
   and for the applewin emulator's rendering as well.

   This should be adequate */

/* 16 Color BMP style palette (Windows XP and Windows 7) */
/* notice that these johnny-come-latelies reversed GRAY and WHITE
   order in their VGA palette index ignoring the default standard
   color index order of the IBM-PC VGA video adapter. */
unsigned char cmapxpbmp[16][4]={
0x00, 0x00, 0x00,VGA_BLACK,
0x00, 0x00, 0x80,VGA_BLUE,
0x00, 0x80, 0x00,VGA_GREEN,
0x00, 0x80, 0x80,VGA_CYAN,
0x80, 0x00, 0x00,VGA_RED,
0x80, 0x00, 0x80,VGA_MAGENTA,
0x80, 0x80, 0x00,VGA_BROWN,
0x80, 0x80, 0x80,VGA_GRAY,
0xC0, 0xC0, 0xC0,VGA_WHITE,
0x00, 0x00, 0xFF,VGA_LBLUE,
0x00, 0xFF, 0x00,VGA_LGREEN,
0x00, 0xFF, 0xFF,VGA_LCYAN,
0xFF, 0x00, 0x00,VGA_LRED,
0xFF, 0x00, 0xFF,VGA_LMAGENTA,
0xFF, 0xFF, 0x00,VGA_YELLOW,
0xFF, 0xFF, 0xFF,VGA_BWHITE};

/* 16 Color BMP style palette (probably PBRUSH from Windows 3.1) */
unsigned char cmapbmp[16][4]={
0x00, 0x00, 0x00,VGA_BLACK,
0x00, 0x00, 0xBF,VGA_BLUE,
0x00, 0xBF, 0x00,VGA_GREEN,
0x00, 0xBF, 0xBF,VGA_CYAN,
0xBF, 0x00, 0x00,VGA_RED,
0xBF, 0x00, 0xBF,VGA_MAGENTA,
0xBF, 0xBF, 0x00,VGA_BROWN,
0xC0, 0xC0, 0xC0,VGA_WHITE,
0x80, 0x80, 0x80,VGA_GRAY,
0x00, 0x00, 0xFF,VGA_LBLUE,
0x00, 0xFF, 0x00,VGA_LGREEN,
0x00, 0xFF, 0xFF,VGA_LCYAN,
0xFF, 0x00, 0x00,VGA_LRED,
0xFF, 0x00, 0xFF,VGA_LMAGENTA,
0xFF, 0xFF, 0x00,VGA_YELLOW,
0xFF, 0xFF, 0xFF,VGA_BWHITE};

/* 16 color ZSOFT PCPAINT PCX style palette
   I can't remember which version of
   PCPaintBrush used this as a default
   or if they all did... */
unsigned char cmappcx[16][4]={
0x00, 0x00, 0x00,VGA_BLACK,
0x00, 0x00, 0xAA,VGA_BLUE,
0x00, 0xAA, 0x00,VGA_GREEN,
0x00, 0xAA, 0xAA,VGA_CYAN,
0xAA, 0x00, 0x00,VGA_RED,
0xAA, 0x00, 0xAA,VGA_MAGENTA,
0xAA, 0xAA, 0x00,VGA_BROWN,
0xAA, 0xAA, 0xAA,VGA_WHITE,
0x55, 0x55, 0x55,VGA_GRAY,
0x55, 0x55, 0xFF,VGA_LBLUE,
0x55, 0xFF, 0x55,VGA_LGREEN,
0x55, 0xFF, 0xFF,VGA_LCYAN,
0xFF, 0x55, 0x55,VGA_LRED,
0xFF, 0x55, 0xFF,VGA_LMAGENTA,
0xFF, 0xFF, 0x55,VGA_YELLOW,
0xFF, 0xFF, 0xFF,VGA_BWHITE};

/* VGA Palette probably converted to 24 bit from actual 18 bit adapter
   settings but I really can't remember */
unsigned char cmapvga[16][4]={
0x00, 0x00, 0x00,VGA_BLACK,
0x00, 0x00, 0xFF,VGA_BLUE,
0x00, 0xFF, 0x00,VGA_GREEN,
0x00, 0xFF, 0xFF,VGA_CYAN,
0xFF, 0x00, 0x00,VGA_RED,
0xFF, 0x00, 0xFF,VGA_MAGENTA,
0xFF, 0xFF, 0x00,VGA_BROWN,
0xC0, 0xC0, 0xC0,VGA_WHITE,
0x55, 0x55, 0x55,VGA_GRAY,
0x55, 0x55, 0xFF,VGA_LBLUE,
0x55, 0xFF, 0x55,VGA_LGREEN,
0x55, 0xFF, 0xFF,VGA_LCYAN,
0xFF, 0x55, 0x55,VGA_LRED,
0xFF, 0x55, 0xFF,VGA_LMAGENTA,
0xFF, 0xFF, 0x55,VGA_YELLOW,
0xFF, 0xFF, 0xFF,VGA_BWHITE};


/* default is applewin rendering - older version */
unsigned char usermap[6][4] = {
0x00, 0x00, 0x00, VGA_BLACK,    /* black */
0x20, 0xC0, 0x00, VGA_GREEN,    /* green */
0xA0, 0x00, 0xFF, VGA_MAGENTA,  /* violet */
0xF0, 0x50, 0x00, VGA_RED,      /* orange */
0x00, 0x80, 0xFF, VGA_BLUE,     /* blue */
0xFF, 0xFF, 0xFE, VGA_WHITE};   /* white */

/* applewin rendering - older version */
unsigned char cmapplewin[6][4] = {
	0,0,0,      VGA_BLACK,    /* black */
	32,192,0,   VGA_GREEN,    /* green */
	160,0,255,  VGA_MAGENTA,  /* violet */
	240,80,0,	VGA_RED,      /* orange */
	0,128,255,  VGA_BLUE,     /* blue */
	255,255,254,VGA_WHITE};   /* white */

/* applewin rendering - newer version */
unsigned char cmapplewin120[6][4] = {
	0,0,0,      VGA_BLACK,    /* black */
	56,203,0,   VGA_GREEN,    /* green */
	199,52,255, VGA_MAGENTA,  /* violet */
	242,94,0,	VGA_RED,      /* orange */
	13,161,255, VGA_BLUE,     /* blue */
	255,255,254,VGA_WHITE};   /* white */

unsigned char cmapplewinfuture[6][4] = {
	0,0,0,      VGA_BLACK,    /* black */
	56,203,0,   VGA_GREEN,    /* green */
	199,52,255, VGA_MAGENTA,  /* violet */
	242,94,0,	VGA_RED,      /* orange */
	13,161,255, VGA_BLUE,     /* blue */
	255,255,254,VGA_WHITE};   /* white */

/* duplicate palettes of the above in applewin color order */
unsigned char cmappleold[6][5] = {
  0,  0,  0,HBLACK, VGA_BLACK,
255,255,254,HWHITE, VGA_WHITE,
  0,128,255,HBLUE,  VGA_BLUE,
240, 80,  0,HORANGE,VGA_RED,
 32,192,  0,HGREEN, VGA_GREEN,
160,  0,255,HVIOLET,VGA_MAGENTA};

unsigned char cmapplenew[6][5] = {
  0,  0,  0,HBLACK,	VGA_BLACK,
255,255,254,HWHITE,	VGA_WHITE,
 13,161,255,HBLUE,	VGA_BLUE,
242, 94,  0,HORANGE,VGA_RED,
 56,203,  0,HGREEN,	VGA_GREEN,
199, 52,255,HVIOLET,VGA_MAGENTA};


/* clipshop rendering */
unsigned char cmappleclip[6][4]={
0x00, 0x00, 0x00, VGA_BLACK,
255,  0x00, 128,  VGA_MAGENTA,
0x00, 192,  0x00, VGA_GREEN,
255,  128,  0x00, VGA_RED,
0x00, 0x00, 192,  VGA_BLUE,
0xFF, 0xFF, 0xFF, VGA_WHITE};

/* GetVgaColor matches BMP colors using the palettes above
   and returns a constant index value, and should always work with 16 color BMP
   files produced by Windows XP Paint and hopefully those other 16 color
   BMP's made elsewhere. */
unsigned char GetVgaColor(unsigned char red, unsigned char green, unsigned char blue)
{

    int idx;

    /* try exact match */
    if (externalmap != 0) {
		for (idx = 0; idx < 6; idx++) {
			if (red == usermap[idx][0] && green == usermap[idx][1] && blue == usermap[idx][2])return usermap[idx][3];
		}
	}

    for (idx = 0; idx < 6; idx++) {
		if (red == cmappleclip[idx][0] && green == cmappleclip[idx][1] && blue == cmappleclip[idx][2])return cmappleclip[idx][3];
	}

    for (idx = 0; idx < 6; idx++) {
		if (red == cmapplewin[idx][0] && green == cmapplewin[idx][1] && blue == cmapplewin[idx][2])return cmapplewin[idx][3];
	}

    for (idx = 0; idx < 6; idx++) {
		if (red == cmapplewin120[idx][0] && green == cmapplewin120[idx][1] && blue == cmapplewin120[idx][2])return cmapplewin120[idx][3];
	}
	if (applewinversion == 911) {
		for (idx = 0; idx < 6; idx++) {
			if (red == cmapplewinfuture[idx][0] && green == cmapplewinfuture[idx][1] && blue == cmapplewinfuture[idx][2])return cmapplewinfuture[idx][3];
		}
	}

    for (idx = 0; idx < 16; idx++) {
		if (red == cmapxpbmp[idx][0] && green == cmapxpbmp[idx][1] && blue == cmapxpbmp[idx][2])return cmapxpbmp[idx][3];
	}
    for (idx = 0; idx < 16; idx++) {
		if (red == cmapbmp[idx][0] && green == cmapbmp[idx][1] && blue == cmapbmp[idx][2])return cmapbmp[idx][3];
	}
    for (idx = 0; idx < 16; idx++) {
		if (red == cmapvga[idx][0] && green == cmapvga[idx][1] && blue == cmapvga[idx][2])return cmapvga[idx][3];
	}
    for (idx = 0; idx < 16; idx++) {
		if (red == cmappcx[idx][0] && green == cmappcx[idx][1] && blue == cmappcx[idx][2])return cmappcx[idx][3];
	}

	/* adjust gun values to match using EGA-like thresholds
	  this corresponds to the old PCX style palette...*/
	if (red < 43) red = 0;         /* 0x00 */
	else if (red < 128) red = 85;  /* 0x55 */
	else if (red < 224) red = 170; /* 0xaa */
	else red = 255;                /* 0xff */

	if (green < 43) green = 0;
	else if (green < 128) green = 85;
	else if (green < 213) green = 170;
	else green = 255;

	if (blue < 43) blue = 0;
	else if (blue < 128) blue = 85;
	else if (blue < 213) blue = 170;
	else blue = 255;

    for (idx = 0; idx < 16; idx++) {
		if (red == cmapvga[idx][0] && green == cmapvga[idx][1] && blue == cmapvga[idx][2])return cmapvga[idx][3];
	}
    for (idx = 0; idx < 16; idx++) {
		if (red == cmappcx[idx][0] && green == cmappcx[idx][1] && blue == cmappcx[idx][2])return cmappcx[idx][3];
	}

	/* that's enough! if no match yet, garbage-in so return 0 - paint-it black */
    return (unsigned char)VGA_BLACK;
}


unsigned char VgaToCGA(unsigned char color)
{

  switch(color) {

	case  VGA_BWHITE:
    case  VGA_WHITE:
    case  VGA_YELLOW:

          color = CGA_WHITE; break;

    case  VGA_LRED:
    case  VGA_RED:
    case  VGA_BROWN:
    case  VGA_GREEN:
    case  VGA_LGREEN:

          /* 2 bit pattern maps to orange or green on the Apple II */
          color = CGA_CYAN; break;

    case  VGA_BLUE:
    case  VGA_LBLUE:
    case  VGA_CYAN:
    case  VGA_LCYAN:
    case  VGA_MAGENTA:
    case  VGA_LMAGENTA:

          /* 2 bit pattern maps to blue or violet on the Apple II */
          color = CGA_MAGENTA; break;

    case  VGA_GRAY:
    case  VGA_BLACK:
    default:
          color = CGA_BLACK; break;
  }
  return color;

}

/* optionally used to set palette bits in an attempt
   to preserve discrete colors when converting
   color bmps to Apple II HGR files... */
unsigned char VgaToApple(unsigned char color)
{

  /* black and white are not used for the
     palette count because they exist in either
     Apple II palette */
  unsigned char pal = HBLACK;

  switch(color) {

    /* the other colors are grouped by apple II palette */
    case  VGA_LRED:
    case  VGA_RED:
    case  VGA_BLUE:
    case  VGA_LBLUE:
    case  VGA_CYAN:
    case  VGA_LCYAN:
          pal = HORANGE; break;

    case  VGA_BROWN:
    case  VGA_GREEN:
    case  VGA_LGREEN:
    case  VGA_MAGENTA:
    case  VGA_LMAGENTA:
          pal = HGREEN;

  }
  return pal;

}

unsigned char VgaToAppleWin(unsigned char color)
{

 switch(color) {

	case  VGA_BWHITE:
    case  VGA_WHITE:
    case  VGA_YELLOW:

          color = HWHITE; break;

    case  VGA_LRED:
    case  VGA_RED:

          if (a2switch == 1) color = HBLUE;
          else color = HORANGE; break;

    case  VGA_BROWN:
    case  VGA_GREEN:
    case  VGA_LGREEN:

          if (a2switch == 1) color = HVIOLET;
          else color = HGREEN; break;

    case  VGA_BLUE:
    case  VGA_LBLUE:
    case  VGA_CYAN:
    case  VGA_LCYAN:

          if (a2switch == 1) color = HORANGE;
          else color = HBLUE; break;

    case  VGA_MAGENTA:
    case  VGA_LMAGENTA:

          /* 2 bit pattern maps to blue or violet on the Apple II */
          if (a2switch == 1) color = HGREEN;
          else color = HVIOLET; break;

    case  VGA_GRAY:
    case  VGA_BLACK:
    default:
          color = HBLACK; break;
  }
  return color;
}

void UnpackVGALine(int width)
{
     int i,j;
     unsigned char c, color;

     j=0;
     for(i=0;i<width;i++)
     {
        c = bmpscanline[i]>>4;
        color = bmpmapinfo[c];
        palettebuf[j] = VgaToApple(color);
        svgabuf[j] = VgaToCGA(color);
        j++;

        c = bmpscanline[i]&0xf;
        color = bmpmapinfo[c];
        palettebuf[j] = VgaToApple(color);
        svgabuf[j] = VgaToCGA(color);
        j++;
     }
}

void UnpackCGALine()
{

/* preserve color patterns on the Apple II as much as possible when
   converting from a GWBASIC BSAVED 4 color Image by expanding the
   colored 2 bit pixel bit pattern */

	unsigned char c;
	int idx, jdx = 0;

    /* unpack the entire line and isolate the colors */
	for (idx = 0; idx < 80; idx++) {

		c = bmpscanline[idx];
		svgabuf[jdx] = (c >> 6) &3; jdx++;
		svgabuf[jdx] = (c >> 4) &3; jdx++;
		svgabuf[jdx] = (c >> 2) &3; jdx++;
		svgabuf[jdx] =  c &3; jdx++;
	}
}


unsigned char cswitch(unsigned char c)
{
	/* swap green and violet option */
	if(c == 2 && a2switch == 1) return 1;
	if(c == 1 && a2switch == 1) return 2;

	return c;
}


void CGAToMono(int xorg)
{

/* preserve color patterns on the Apple II as much as possible when
   converting from a GWBASIC BSAVED 4 color Image by packing every
   second colored 2 bit pixel bit pattern into an 8 bit
   "monochrome" line. This process is a lossy general purpose
   algorithm for sure. However on a 4 color CGA Image with line
   elements of 2 colors in width, results are reasonable and
   overall loss of detail is minimized despite half-scaling to
   the 140 pixel Apple II format. */

     unsigned char c, byte;
	 int idx, jdx = 0;

     for (idx = 0; idx < 35; idx++) {

	   /* skip every second pixel */
	   c = cswitch(svgabuf[xorg]);
	   byte = (c << 6);xorg+=2;

	   c = cswitch(svgabuf[xorg]);
	   byte |= (c << 4);xorg+=2;

	   c = cswitch(svgabuf[xorg]);
	   byte |= (c << 2);xorg+=2;

	   c = cswitch(svgabuf[xorg]);
	   byte |= c;xorg+=2;

	   bmpscanline[idx] = byte;
	}

}

void setVGAPalette(int xorg)
{

     /* this algorithm is very flawed in two major respects.

        the first and greatest flaw is the nature of how
        discrete colors on the apple II itself are controlled
        by the palette bit settings and there is nothing that
        can be done about this of course.

        the second flaw is that I am not looking ahead into the
        next pixel block to determine what follows except in the
        case of a 3.5 pixel block when xPalette is set to 7.

        However, in my view this program is complicated enough
        to use without introducing the level of complexity
        that would be required for further control of
        precedence by the user. Even in my Bmp2Rag utility
        which has a number of strategies that are more set-based
        I stopped-short of adding additional filters for this
        purpose pretty-much for the same reasons of complexity of use.

        */

     unsigned char c, oldbit, newbit;
	 int idx, jdx = 0, cnt = 0, orange=0, green=0, bitcnt=0, val,val2;

     /* if the xPalette option is not in use we leave
        our palettebits alone ... */
     clearpalettebits();
     if (xPalette != 7 && xPalette != 14) return;


     /* I am only concerning myself with pixels
        that will survive in the output image so I
        use a variation of the same algorithm that
        packs colored scanlines into monochrome lines
        in CGAToMono */

     oldbit = palettebits[0]; /* seed forward */

     for (idx = 0; idx < 35; idx++) {

	   /* skip every second pixel */
	   for (jdx = 0; jdx < 4; jdx++) {
	   		c = palettebuf[xorg];
	   		xorg+=2;

	        /* single byte option */
	        if (xPalette == 7) {

				/* weighted value of split pixels is 50%
				   of a whole pixel */
				val  = 2;
				val2 = 0;
				if (cnt%7 == 3)val = val2 = 1;
	   			if (c == HORANGE)orange+=val;
	   			if (c == HGREEN)green+=val;

                if (cnt%7 == 3 || cnt %7 == 6) {
					newbit = oldbit;
				    if (green != 0 || orange != 0) {
						if (green > orange) newbit = 0;
						if (orange > green) newbit = 0x80;
					}
					palettebits[bitcnt] = newbit;
					bitcnt++;
					orange = 0;
					green = 0;
					/* carry forward split pixels */
					if (c == HORANGE)orange = val2;
					if (c == HGREEN)green = val2;
					oldbit = newbit; /* carry forward */
				}

			}
			else {
                /* paired byte option */
	   			if (c == HORANGE)orange++;
	   			if (c == HGREEN)green++;

                if (cnt%7 == 6) {
				    newbit = oldbit;
				    if (green != 0 || orange != 0) {
						if (green > orange) newbit = 0;
						if (orange > green) newbit = 0x80;
					}
					orange = 0;
					green = 0;
					palettebits[bitcnt] = newbit; bitcnt++;
					palettebits[bitcnt] = newbit; bitcnt++;
					oldbit = newbit; /* carry forward */
				}

			}
			cnt++;
	   }
	}
}


/* read GWBASIC 320 x 200 BSAVED Image */
int BAS_Read(char *basename, int x, int y)
{
    int fh,fh2,y1,i;
    unsigned char name[128], headbuf[7];

    sprintf(name,"%s.BAS",basename);

    /* open the file twice... interleaved */

    if((fh = open(name,O_RDONLY|O_BINARY)) == -1) return INVALID;
    if((fh2 = open(name,O_RDONLY|O_BINARY)) == -1) {
		close(fh);
		return INVALID;
	}

    if (read(fh,headbuf,7) != 7) {
		close(fh);
		close(fh2);
		return INVALID;
	}

    /* I don't check the file length...
       it sometimes varies slightly... */
    for(i=0;i!=4;i++){
		if(headbuf[i]!=BSAVED_header[i]){
        	close(fh)  ;
        	close(fh2) ;
            return INVALID  ;
        }
	}

    lseek(fh,(long)7,SEEK_SET)  ;
    lseek(fh2,(long)8199,SEEK_SET);

    if (y > 8)y = 8;
    if (y < 0)y = 0;
    /* preserve 2 bit patterns */
    if (x%2 != a2shift) x--;
    while(x > 40) x-=2;

    /* 2 bit patterns of CGA magenta and cyan palette correspond
       to blue and orange on the apple II in palette 1 so
       swap them to maintain similar colors. */
    if (a2switch == 1) a2switch = 0;
    else a2switch = 1;

    /* the following windows a 280 x 192 portion
       of a GWBASIC CGA 320 x 200 x 4-Color BSAVED Image
       and translates the CGA 2 bit color patterns into
       a half-scaled monochrome line.

       The monochrome line is then passed into
       the same core monochrome conversion functions
       that build the Apple II hgr output from monochrome
       BMP files.

       a neat trick if a guy knows how to do it.

       */

	memset(&hgrbuf[0],0,8192); /* clear write buffer */
	clearpalettebits();

    i = 0;
    for(y1=0;y1<200;y1++){
	   if (y1%2 == 0)read(fh,bmpscanline,80);
	   else read(fh2,bmpscanline,80);
	   if (y1 < y) continue;
	   /* convert the cga line to a mono line */
       UnpackCGALine();
       CGAToMono(x);

 	   /* pass the converted line to the core
 	      Apple II conversion functions */
 	   ibmbits();    // decode 8 bit
	   applebits(i); // encode 7 bit
       i++;
       if (i> 191) break;
    }

	close(fh);
	close(fh2);
	return SUCCESS;

}

int BMP_Read(char *basename, int x, int bot)
{

    FILE *fp;
    int packet = INVALID, y1,vres,i;
    char bmpfile[128];

	sprintf(bmpfile,"%s.BMP",basename);
    if((fp=fopen(bmpfile,"rb"))==NULL)return INVALID;

    /* read the header stuff into the appropriate structures */
    fread((char *)&BitMapFileHeader.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&bmp.bmiHeader.biSize,
                 sizeof(VGABITMAPINFO),1,fp);

    if (bmp.bmiHeader.biCompression==BI_RGB &&
        BitMapFileHeader.bfType[0] == 'B' && BitMapFileHeader.bfType[1] == 'M' &&
        bmp.bmiHeader.biPlanes==1  && bmp.bmiHeader.biBitCount==4) {

    	if (bmp.bmiHeader.biWidth == 320 && bmp.bmiHeader.biHeight == 200) {
			packet = 160;
			vres = 200;
    		if (bot > 8)bot = 8;
    		if (bot < 0)bot = 0;
    		if (x%2 != a2shift) x--; /* preserve 2 bit patterns */
    		while(x > 40) x-=2;
		}
    	if (bmp.bmiHeader.biWidth == 280 && bmp.bmiHeader.biHeight == 192) {
			packet = 140;
			vres = 192;
			x = 0;
			bot = 0;
		}

	}

    if(packet == INVALID){ fclose(fp);return INVALID;}

    /* map the colors in the bmp to a standard index
       value suitable for conversion
       this should work flawlessly for bmps created in
       Windows Paint for Windows XP or Windows 7. */
    for(i=0;i<16;i++) bmpmapinfo[i] =
     GetVgaColor(bmp.bmiColors[i].rgbRed,
                 bmp.bmiColors[i].rgbGreen,
                 bmp.bmiColors[i].rgbBlue);

	memset(&hgrbuf[0],0,8192); /* clear write buffer */

    bot += 192;
    i = 191; /* bmp's are upside-down
                so conversion of scanlines is in
                reverse order */
    for(y1=vres;y1>0;y1--)
    {
	  fread((char *)&bmpscanline[0],1,packet,fp);
	  if (y1 > bot) continue;
	  UnpackVGALine(packet);
	  setVGAPalette(x);
      CGAToMono(x);
 	  /* pass the converted line to the core
 	     Apple II conversion functions */
 	  ibmbits();    // decode 8 bit
	  applebits(i); // encode 7 bit
      i--;
      if (i< 0) break;
    }
    fclose(fp);

return SUCCESS;

}

/* read apple II hgr input file */
int read_bin(unsigned char *name)
{

	FILE *fp;

  	fp = fopen(name,"rb");
	if (NULL == fp)return INVALID;
	fread(hgrbuf,1,8192,fp);
	fclose(fp);

	return SUCCESS;
}

/* read apple II hgr bot, top or rag file */
/* input is raw only... maybe next version will
   support RAX format if any need at all...
   debatable whether anyone will use compressed
   monochrome images much or if a busy image will
   compress much anyway... */
int read_bot(unsigned char *name)
{
	FILE *fp;
	unsigned char width = 0, height = 0, *ptr;
    int x, y, xoff;


	fp = fopen(name,"rb");
	if (NULL == fp) return INVALID;

    /* width in bytes x height in rasters */
    width =  fgetc(fp);
    height = fgetc(fp);

    /* support for fullscreen images only */
    if (width != 40 || (height != 192 && height != 160)) {
		fclose(fp);
	    return INVALID;
	}

    memset(&hgrbuf[0],0,8192);
    for (y = 0; y < height; y++) {
	    xoff = HB[y] - 0x2000;
    	ptr  = (unsigned char *) &hgrbuf[xoff];
		fread(ptr,1,40,fp);
	}
	fclose(fp);

	return SUCCESS;

}

/* decodes apple II hgr buffer */
int applebites(int y)
{
	    int xoff,idx;
	    unsigned char *ptr, ch, pal;

	    xoff = HB[y];
	    ptr  = (unsigned char *) &hgrbuf[xoff-0x2000];

        xoff = 0;
        for (idx = 0; idx < 40; idx++) {

            ch = ptr[idx];
            pal = (ch >> 7)&1;

            buf280[xoff] = ((ch) &1); pal280[xoff] = pal; xoff++;
            buf280[xoff] = ((ch >> 1) &1); pal280[xoff] = pal; xoff++;
            buf280[xoff] = ((ch >> 2) &1); pal280[xoff] = pal; xoff++;
            buf280[xoff] = ((ch >> 3) &1); pal280[xoff] = pal; xoff++;
            buf280[xoff] = ((ch >> 4) &1); pal280[xoff] = pal; xoff++;
            buf280[xoff] = ((ch >> 5) &1); pal280[xoff] = pal; xoff++;
            buf280[xoff] = ((ch >> 6) &1); pal280[xoff] = pal; xoff++;

 		}
		return SUCCESS;

}



/* encodes bmp scanline */
int ibmbites()
{
     int i,j,k;
     unsigned char bits[8], ch;

     j=0;
     for(i=0;i<35;i++)
     {
        ch = 0;
        for(k=0;k<8;k++)
        {
		  bits[k] = buf280[j]; j++;
		}
		bmpscanline[i] = (bits[0]<<7|bits[1]<<6|bits[2]<<5|bits[3]<<4|
                          bits[4]<<3|bits[5]<<2|bits[6]<<1|bits[7]);
     }
     return SUCCESS;

}

/* save a monochrome bmp file */
int save_to_bmp(unsigned char *basename)
{

    FILE *fp;
    unsigned char outfile[128];
	int y,y2;

    sprintf(outfile,"%s.BMP",basename);
	fp = fopen(outfile,"wb");
	if (NULL == fp)return INVALID;

    /* write header */

    fwrite(mono192,1,sizeof(mono192),fp);

    /* write scanlines */

    y2 = 191;
    bmpscanline[35] = 0xff; // padding
   	for (y = 0; y< 192; y++) {
       applebites(y2);
       ibmbites();
       fwrite(bmpscanline,1,36,fp);
	   y2 -= 1;
    }

    return SUCCESS;

}


/* encodes 16 color bmp scanline
   using color-pair strategy */
int ibmbites16()
{
     int i,j,k;
     unsigned char b1, b2, c1, c2;


     for(i=0,j=1,k =0;i<280;i+=2,j+=2,k++)
     {
		b1 = buf280[i];
		b2 = buf280[j];

		if (b1 == 0 && b2 == 0) {
		   bmpscanline[k] = 0;
		   continue;
		}
		if (b1 == 1 && b2 == 1) {
		   bmpscanline[k] = 255;
		   continue;
		}

		if (b1 == 0 && b2 == 1) {
		    if (pal280[i] == 1) c1 = VGA_LBLUE;
		    else c1 = VGA_MAGENTA;

		    if (pal280[j] == 1) c2 = VGA_LBLUE;
		    else c2 = VGA_MAGENTA;
		}

        if (b1 == 1 && b2 == 0) {
		    if (pal280[i] == 1) c1 = VGA_LRED;
		    else c1 = VGA_LGREEN;

		    if (pal280[j] == 1) c2 = VGA_LRED;
		    else c2 = VGA_LGREEN;

		}


		bmpscanline[k] = (c1 << 4 | c2);

     }
     return SUCCESS;

}


/* save a 16 color bmp file */
int save_to_bmp16(unsigned char *basename)
{

    FILE *fp;
    unsigned char outfile[128];
	int y,y2;

    sprintf(outfile,"%s.BMP",basename);
	fp = fopen(outfile,"wb");
	if (NULL == fp)return INVALID;

    /* write header */
    fwrite(color280,1,sizeof(color280),fp);

    /* write scanlines */

    y2 = 191;
   	for (y = 0; y< 192; y++) {
       applebites(y2);
       ibmbites16();
       fwrite(bmpscanline,1,140,fp);
	   y2 -= 1;
    }

    fclose(fp);
    return SUCCESS;

}

/* if we have a 24 bit image or an svga image, assume it is an AppleWin Screen Grab */
unsigned char applewinindex(unsigned char r, unsigned char g, unsigned char b)
{
	unsigned char idx;

    idx = GetVgaColor(r,g,b);
    return VgaToAppleWin(idx);
}

/* rgb triples from applewin screen grab using windows clipboard
   or svga indices from applewin screen save */
unsigned char AppleWinLine[1680];


/* converting from applewin bmp's back to hgr should be accurate
   since all we are doing is reversing applewin's  rendering */
void UnpackAppleWinLine(int packet)
{
     int i,j,k, green, orange;
     unsigned char r, g, b, p;

     /* convert rgb triples or svga indices to HGR colors */
     for(i=0,j=0;i<280;i++) {
		 /* 24 bit */
		 if (packet == 840 || packet == 1680) {
		 	b =  AppleWinLine[j]; j++;
		 	g =  AppleWinLine[j]; j++;
		 	r =  AppleWinLine[j]; j++;
		 }
		 /* 8 bit */
		 if (packet == 280 || packet == 560) {
			 p = AppleWinLine[j]; j++;
			 r = sbmp[p].rgbRed;
             g = sbmp[p].rgbGreen;
             b = sbmp[p].rgbBlue;
		 }
		 AppleWinLine[i] = applewinindex(r,g,b);
		 /* skip every second triple if the image is double-scale */
		 if (packet == 1680) j+= 3;
		 /* skip every second indices if the image is double scaled */
		 if (packet == 560) j++;
	 }

     /* create bit pattern from pixelized values */
     /* i is even and j is odd */
	 for (i= 0, j = 1; i < 280; i+=2, j+=2) {
	    buf280[i] = 0; /* assume everything is black */
	    buf280[j] = 0;

        /* add the white bits - this also accounts for the half shift of
           the color pixels which applewin renders as white to represent
           aliasing of the color anomalies */
	    if (AppleWinLine[i] == HWHITE) buf280[i] = 1;
	    if (AppleWinLine[j] == HWHITE) buf280[j] = 1;

        /* add the violet or blue bits - the 2-bit value will be 2 */
	    if (AppleWinLine[i] == HBLUE || AppleWinLine[i] == HVIOLET )buf280[i] = 1;
	    if (AppleWinLine[j] == HBLUE || AppleWinLine[j] == HVIOLET )buf280[j] = 0;

        /* add the green or orange bits - the 2-bit value will be 1 */
	    if (AppleWinLine[i] == HORANGE || AppleWinLine[i] == HGREEN )buf280[i] = 0;
	    if (AppleWinLine[j] == HORANGE || AppleWinLine[j] == HGREEN )buf280[j] = 1;

	 }

     /* set the palette based on groups of seven pixels */
	 clearpalettebits();
	 p = palettebits[0];
	 for (i = 0, k=0; i < 40; i++ ) {
		 orange = 0;
		 green = 0;
		 for (j = 0; j < 7; j++) {
			 if (AppleWinLine[k] == HORANGE ||  AppleWinLine[k] == HBLUE) orange++;
			 if (AppleWinLine[k] == HGREEN ||  AppleWinLine[k] == HVIOLET) green++;
			 k++;
		 }
		 if (green > orange) p = 0;
		 if (orange > green) p = 0X80;

		 palettebits[i] = p;
	 }

}

/* 8 bit decoder only */
/* I separated the AppleWin Code for the newer version from the code for the
   older version in case I need to tweak this further...
   for now the only change is to advance by 1 byte into the scanline
   which will preserve orange and violet both */
void UnpackAppleWin120Line(int packet)
{
     int i,j=0,k, green, orange;
     unsigned char r, g, b, p;

     if (packet == 560) j++;

     /* convert rgb triples or svga indices to HGR colors */
     for(i=0;i<280;i++) {
	     /* 8 bit */
		 if (packet == 280 || packet == 560) {
			 p = AppleWinLine[j]; j++;
			 r = sbmp[p].rgbRed;
             g = sbmp[p].rgbGreen;
             b = sbmp[p].rgbBlue;
		 }
		 AppleWinLine[i] = applewinindex(r,g,b);
		 /* skip every second indices if the image is double scaled */
		 if (packet == 560) j++;
	 }

     /* create bit pattern from pixelized values */
     /* i is even and j is odd */
	 for (i= 0, j = 1; i < 280; i+=2, j+=2) {
	    buf280[i] = 0; /* assume everything is black */
	    buf280[j] = 0;

        /* add the white bits - this also accounts for the half shift of
           the color pixels which applewin renders as white to represent
           aliasing of the color anomalies */
	    if (AppleWinLine[i] == HWHITE) buf280[i] = 1;
	    if (AppleWinLine[j] == HWHITE) buf280[j] = 1;

        /* add the violet or blue bits - the 2-bit value will be 2 */
	    if (AppleWinLine[i] == HBLUE || AppleWinLine[i] == HVIOLET )buf280[i] = 1;
	    if (AppleWinLine[j] == HBLUE || AppleWinLine[j] == HVIOLET )buf280[j] = 0;

        /* add the green or orange bits - the 2-bit value will be 1 */
	    if (AppleWinLine[i] == HORANGE || AppleWinLine[i] == HGREEN )buf280[i] = 0;
	    if (AppleWinLine[j] == HORANGE || AppleWinLine[j] == HGREEN )buf280[j] = 1;

	 }

     /* set the palette based on groups of seven pixels */
	 clearpalettebits();
	 p = palettebits[0];
	 for (i = 0, k=0; i < 40; i++ ) {
		 orange = 0;
		 green = 0;
		 for (j = 0; j < 7; j++) {
			 if (AppleWinLine[k] == HORANGE ||  AppleWinLine[k] == HBLUE) orange++;
			 if (AppleWinLine[k] == HGREEN ||  AppleWinLine[k] == HVIOLET) green++;
			 k++;
		 }
		 if (green > orange) p = 0;
		 if (orange > green) p = 0X80;

		 palettebits[i] = p;
	 }

}



int AppleWinToHGR(unsigned char *basename)
{

 	FILE *fp;
    int packet = INVALID, y,i,j;
    char bmpfile[128];

	sprintf(bmpfile,"%s.BMP",basename);
    if((fp=fopen(bmpfile,"rb"))==NULL) {
		return INVALID;
	}

    /* read the header stuff into the appropriate structures */
    fread((char *)&BitMapFileHeader.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&bmi.biSize,
                 sizeof(BITMAPINFOHEADER),1,fp);

    if (bmi.biCompression==BI_RGB &&
        BitMapFileHeader.bfType[0] == 'B' && BitMapFileHeader.bfType[1] == 'M' &&
        bmi.biPlanes==1) {

		if (bmi.biBitCount == 24) {
			if (bmi.biWidth == 280 && bmi.biHeight == 192) {
				packet = 840;
			}
			if (bmi.biWidth == 560 && bmi.biHeight == 384) {
				packet = 1680;
			}
	    }
		if (bmi.biBitCount == 8) {
			if (bmi.biWidth == 280 && bmi.biHeight == 192) {
				packet = bmi.biWidth;
			}
			if (bmi.biWidth == 560 && bmi.biHeight == 384) {
				packet = bmi.biWidth;
			}

			if (packet != INVALID) {
				applewinversion = 120;
				fread((char *)&sbmp[0].rgbBlue, sizeof(RGBQUAD)*256,1,fp);

				/* check applewin version - some of this for future versions */
                for (i = 16,j=0; j < 6; i++,j++) {
					 applewinversion = 0;
			         if (sbmp[i].rgbRed != cmapplenew[j][0]) break;
				     if (sbmp[i].rgbGreen != cmapplenew[j][1]) break;
                     if (sbmp[i].rgbBlue != cmapplenew[j][2]) break;
                     applewinversion = 120;
				}
				if (applewinversion == 0) {
					for (i = 16,j=0; j < 6; i++,j++) {
						 applewinversion = 0;
						 if (sbmp[i].rgbRed != cmappleold[j][0]) break;
						 if (sbmp[i].rgbGreen != cmappleold[j][1]) break;
						 if (sbmp[i].rgbBlue != cmappleold[j][2]) break;
						 applewinversion = 116;
					}
			 	}
				if (applewindex == 1 && applewinversion == 0) {
					applewinversion = 911;
					for (i = 16,j=0; j < 6; i++,j++) {
				       	y = cmapplenew[j][3];
				       	cmapplewinfuture[y][0] = sbmp[i].rgbRed;
				       	cmapplewinfuture[y][1] = sbmp[i].rgbGreen;
				       	cmapplewinfuture[y][2] = sbmp[i].rgbBlue;
				       	cmapplewinfuture[y][3] = cmapplenew[j][4];
					}
				}

		    }
	    }
	}

    if(packet == INVALID){ fclose(fp);return INVALID;}

	memset(&hgrbuf[0],0,8192); /* clear write buffer */


    i = 191; /* bmp's are upside-down
                so conversion of scanlines is in
                reverse order */
    for(y=192;y>0;y--)
    {
	  fread((char *)&AppleWinLine[0],1,packet,fp);
	  /* skip every second scanline if the image is double-scale */
	  if (packet == 1680 || packet == 560) fread((char *)&AppleWinLine[0],1,packet,fp);
	  if (applewinversion==120) UnpackAppleWin120Line(packet);
	  else UnpackAppleWinLine(packet);
	  applebits(i); // encode 7 bit
      i--;
      if (i< 0) break;
    }
    fclose(fp);

return SUCCESS;

}



int ReadExternalMapFile(char *mapfile)
{

	unsigned char *ptr, buf[128];
	FILE *fp;
	int r,g,b,i = 0,j;

	sprintf(buf,"%s.BAP",mapfile);

    /* user defined color option */
    if((fp=fopen(buf,"r"))!=NULL) {
	  while (i<6) {
		if (NULL == fgets(buf,80,fp)) break;
		if (buf[0] < 48 || buf[0] > 57) continue;
		ptr = (char *)&buf[0];
		r = atoi(ptr);
		if (r < 0 || r > 255) continue;
		g = -1; b = -1;
		for (j=0;buf[j]!=0;j++) {
			if (buf[j] == ',') {
				ptr = (char *)&buf[j+1];
				if (g == -1) g = atoi(ptr);
				else b = atoi(ptr);
				if (g < 0||g > 255) break;

			}
			if (b!=-1) {
				if (b < 0||b > 255) break;
				usermap[i][0] = (unsigned char)r;
				usermap[i][1] = (unsigned char)g;
				usermap[i][2] = (unsigned char)b;
				i++;
				break;
			}
		}
	  }
	  fclose(fp);
	}
	return i;

}


int ReadAppleWin(unsigned char *fullname, unsigned char *basename)
{

 	FILE *fp;
    int packet = INVALID, y,i,j,k;
    char bmpfile[128], outfile[128];

    /* if file does not exist just exit and save the
       other routines the trouble... */
    if((fp=fopen(fullname,"rb"))==NULL) {
		    printf("%s cannot be opened.\n", fullname);
		    exit(1);
	}

    /* if not a bmp file then let the other routines take over */
    sprintf(bmpfile,"%s.BMP",basename);
    if((fp=fopen(bmpfile,"rb"))==NULL) {
		return INVALID;
	}

    /* read the header stuff into the appropriate structures,
       it's likely a bmp file */
    memset(&BitMapFileHeader.bfType,0,sizeof(BITMAPFILEHEADER));
    memset(&bmi.biSize,0,sizeof(BITMAPINFOHEADER));

    fread((char *)&BitMapFileHeader.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&bmi.biSize,
                 sizeof(BITMAPINFOHEADER),1,fp);

    /* check for an applewin bmp */
    if (bmi.biCompression==BI_RGB &&
        BitMapFileHeader.bfType[0] == 'B' && BitMapFileHeader.bfType[1] == 'M' &&
        bmi.biPlanes==1 && bmi.biBitCount == 8) {

        if ((bmi.biWidth == 560 &&  bmi.biHeight == 384) || (bmi.biWidth == 280 &&  bmi.biHeight == 192))  {
		    fread((char *)&sbmp[0].rgbBlue, sizeof(RGBQUAD)*256,1,fp);
		    if (sbmp[17].rgbRed == 255 && sbmp[17].rgbGreen == 255 && sbmp[17].rgbBlue == 254) packet = (int) bmi.biWidth;
		}
	}

    if(packet == INVALID){
		printf("%s is not a Supported AppleWin Screen Capture.\n", bmpfile);
		fclose(fp);
		exit(1);
	}

	memset(&hgrbuf[0],0,8192); /* clear write buffer */
    clearpalettebits();

    i = 191; /* bmp's are upside-down
                so conversion of scanlines is in
                reverse order */
    memset(&AppleWinLine[0],0,562);
    for(y=192;y>0;y--)
    {
	  memset(&buf280[0],0,280);
	  fread((char *)&AppleWinLine[0],1,packet,fp);
	  if (packet == 560) {
		  /* skip every second scanline  */
		  fread((char *)&AppleWinLine[0],1,packet,fp);
		  for (j=(a2shift*2),k=0;k<280;j+=2,k++) {
			if (AppleWinLine[j] != 0 && AppleWinLine[j] != 16)buf280[k] = 1;
		  }
	  }
	  else {
		  for (j=a2shift,k=0;k<280;j++,k++) {
			if (AppleWinLine[j] != 0 && AppleWinLine[j] != 16)buf280[k] = 1;
		  }
	  }
	  applebits(i); // encode 7 bit mono line
      i--;
      if (i< 0) break;
    }
    fclose(fp);

    savehgrfiles(basename);

	sprintf(outfile,"%s.BM2",basename);
	remove(outfile);
	rename(bmpfile,outfile);

return SUCCESS;

}

void main(int argc, char **argv)
{
  int status = 0, idx, bmp = 0, rag = 0,
                       bin = 0, bas = 0,
                       x = 0, y = 0, tempnum;

  unsigned char fname[128],sname[128],outfile[128], c, d, e;
  unsigned char *wordptr;
  unsigned char scratchbuf[128];

  puts(szTextTitle);
  if(argc == 1) {
	puts("For Windows BMP Output...");
    puts("  Command line Usage is \"HMON MyAppleIIFile\"");
    puts("  Apple II Filetypes      BIN, BOT, TOP, RAG");
    puts("  Color Output Option:  \"HMON MyHires.BIN 16\" (example)");
    puts("  AppleWin Mono Screen Capture Input Option:");
    puts("                        \"HMON MyAppleWinMono.BMP 256\"");
    puts("Output BMP is 280 x 192 MyMono.bmp (default) or My16Color.bmp");
    puts("For Apple II Output...(* = Options do not apply to Mono)");
    puts("  Command line Usage is \"HMON My.BMP\" or \"HMON My.BAS\"");
    puts("  Options:   Green Violet Palette: \"HMON My.BMP G\"");
    puts(" (partial)   Reverse Video:        \"HMON My.BMP R\"");
    puts("           * Switch Colors:        \"HMON My.BMP C\"");
    puts("             Shift Pixels:         \"HMON My.BMP S\"");
    puts("             Center Image:         \"HMON My.BMP M\"");
    puts("             Image Window:         \"HMON My.BMP X= Y=\"");
    puts("  Output 280 x 160 TOP file:       \"HMON My.BMP T\"");
    puts("  Some Options: \"HMON My.BMP G R C S M T X= Y= P\" (P14,P7)");
    puts("  AppleWin Future Versions: new palette: \"HMON My.BMP AWIN\"");
    puts("3 Character File extensions must match!");
    puts("8 Character Basename Maximum (MS-DOS Filenames 8.3 Naming Only!");
    puts("  Apple II \"raw full and mixed-screen formats\" only!");
    printf("Enter Input FileName (Blank to Exit): ");
    gets(fname);
    if (fname[0] < 33)
      exit(1);
  }
  else {
    strcpy(fname, argv[1]);

    /* get opts */
    for (idx = 2; idx < argc; idx++) {
	   wordptr = (unsigned char *)&argv[idx][0];
	   if (c == '-') wordptr = (unsigned char *)&argv[idx][1];
	   tempnum = atoi(wordptr);
	   if (tempnum == 16) {
	       a216 = 1;
	       continue;
	   }
	   if (tempnum == 256) {
	       a256 = 1;
	       continue;
	   }
       c = toupper(wordptr[0]);
       switch (c) {
		   case 'A': applewindex = 1; break;
           case 'T': a2top = 1; break;
           case 'G':
           case 'V': a2palette = 0; break;
           case 'R': a2reverse = 1; break;
           case 'C': a2switch = 1; break;
           case 'S': a2shift = 1; break;
		   case 'M':
				   /* center clipping area - default is top left */
				   a2center = 1;
				   y = 4;
				   x = 20;
		   case 'X':
		   case 'Y':
		   			if (wordptr[1] != '=') break;
				   /* x= or y= options for conversion of 320 x 200 bmp's and
										   GWBASIC BSaved Images */
				   tempnum = atoi(&wordptr[2]);
				   if (tempnum > 0) {
					  if (c == 'X' && tempnum > 0 && tempnum < 41) x = tempnum;
					  if (c == 'Y' && tempnum > 0 && tempnum < 9)  y = tempnum;
				   }
		   		   break;
		   case 'P':
				   /* this option is added to the global option when a color bmp
					   is output to an Apple II HGR file... */
				   tempnum = atoi(&wordptr[1]);
				   if (tempnum == 7 || tempnum == 14)xPalette = tempnum;
				   break;
		}

	}
	if (argc > 2)
	  printf("Some Options: G=%d, R=%d, C=%d, S=%d, M=%d, T=%d\n"
	         "              X=%d, Y=%d, P=%d 16=%d\n",
	          a2palette, a2reverse, a2switch, a2shift,a2center, a2top,
	          x,y,xPalette,a216);
  }

  c = d = e = 32;

  for (idx = 0; fname[idx] != ASCIIZ; idx++) {

      if (fname[idx] == '.') {
		  c = toupper(fname[idx + 1]);
		  d = toupper(fname[idx + 2]);
		  e = toupper(fname[idx + 3]);
		  /* support for our extensions only */
		  if (c == 'B' && d == 'A' && e == 'S') {
			  bas = 1;
			  bmp = 1;
		  }
		  if (c == 'B' && d == 'M' && e == 'P') bmp = 1;
		  if (c == 'B' && d == 'I' && e == 'N') bin = 1;
		  if (c == 'B' && d == 'O' && e == 'T') rag = 1;
		  if (c == 'T' && d == 'O' && e == 'P') rag = 1;
		  if (c == 'R' && d == 'A' && e == 'G') rag = 1;

		  break;
	  }

  }

  strcpy(sname, fname);
  wordptr = strtok(sname, ".");
  strcpy(outfile,sname);


  switch(bmp)
  {
	  case 1:   if (a256 == 0) {
		            if (bas == 1) {
						status = BAS_Read(sname,x,y);
					}
					else {
						status = BMPMono_Read(sname,x,y);
						if (status!=0) externalmap = ReadExternalMapFile(sname);
						if (status!=0) status = BMP_Read(sname,x,y);
						if (status!=0) status = AppleWinToHGR(sname);

					}

					if (status) {
					  puts(szTextTitle);
					  printf("\n%s is an Unsupported Format or cannot be opened.\n", fname);
					  exit(1);
					}
					status = savehgrfiles(outfile);
					if (status == SUCCESS)printf("\n%s.BIN and %s.BOT Saved!\n",outfile,outfile);
					else printf("\nError saving %s.BIN and %s.BOT!\n",outfile,outfile);
					break;
				}
				status = ReadAppleWin(fname,sname);

	  default: if (bin == 0 && rag == 0 && bmp == 0) {
				  puts(szTextTitle);
				  printf("\n%s is an Unsupported Filename.\nFile extension %c%c%c is not supported.\n", fname, c,d,e);
				  exit(1);
				}
				if (bin == 1) status = read_bin(fname);
				if (rag == 1) status = read_bot(fname);
				if (status) {
				  puts(szTextTitle);
				  printf("\n%s is an Unsupported Format or cannot be opened.\n", fname);
				  exit(1);
				}

                if (a216 == 0)
	              status = save_to_bmp(outfile);
	            else
	              status = save_to_bmp16(outfile);

	   			if (status == SUCCESS)printf("\n%s.BMP Saved!\n",outfile);
	   			else printf("\nError saving %s.BMP!\n",outfile);
	            break;

  }

  puts("Have a Nice Dos!");
  exit(0);

}
