/* ------------------------------------------------------------------------ */
/* RAXPACK.C (C) Copyright Bill Buckels 2013                                */
/* All Rights Reserved.                                                     */
/*                                                                          */
/* Licence Agreement                                                        */
/* -----------------                                                        */
/*                                                                          */
/* You have a royalty-free right to use, modify, reproduce and              */
/* distribute this source code in any way you find useful,                  */
/* provided that you agree that Bill Buckels has no warranty obligations    */
/* or liability resulting from said distribution in any way whatsoever.     */
/* If you don't agree, remove this source code from your computer now.      */
/*                                                                          */
/* Written by   : Bill Buckels                                              */
/* Email:         bbuckels@escape.ca                                        */
/*                                                                          */
/* Purpose      : This utility will allow you to convert from 				*/
/*                Apple II Hi-Res BSaved (.BIN) Images to                   */
/*				  packed equivalent Apple II Hi-Res images.                 */
/*
					Output:

                      RAX (Hi-Res Xpacked) images (fullscreen).
                      BOT (Hi-Res Raster) images (fullscreen).
                      TOP (Hi-Res Raster) images (280 x 160).

	The RAX is a raster based image of Apple II HGR data encoded as a
	chunk using ZSoft PCX run length encoding. The space saving will vary
	and in some cases it may not be worth-it. In this program I am only
	supporting a fullscreen conversion to RAX.

	The BOT is the non-encoded equivalent and is 7682 bytes which offers
	a modest saving over the 8192 bytes of a BSaved Image. The TOP is
	provides the same data as the BOT until the last 32 scanlines.
	It ends at scanline 159 above the 4 text lines in the Apple II's
	mixed-text and graphics mode. It is basically a truncated BOT.
	At 6402 bytes it is useful for saving a little more disk space
	if programs that run with mixed text and graphics don't need graphics
	in that area.

	I should note that the BOT and TOP are in the same format as
	my RAG image fragment format but as a matter of convenience I have
	always separated them and will likely continue to do so. But
	they use the same loader in my Aztec C programs.

																			*/
/* Revision     : 1.0 First Release                                         */
/* ------------------------------------------------------------------------ */
/* Written in Large Model 16 bit Microsoft C (MSC) Version 8.00c            */
/* Note: Run in an MS-DOS emulator like DOSBox if you can't run it raw.     */
/* ------------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ------------------------------------------------------------------------ */
/* Declarations, Vars. etc.                                                 */
/* ------------------------------------------------------------------------ */


#define ASCIIZ 0

unsigned char *szTextTitle =
    "RAXPACK(C) Version 1.0 Copyright Bill Buckels 2013\n"
    "All Rights Reserved.";

#define SUCCESS  0
#define VALID    SUCCESS
#define FAILURE  -1
#define INVALID  FAILURE


/* Apple 2 Hires Format */

/* provides base address for page1 hires scanlines  */
unsigned HB[]={
0x2000, 0x2400, 0x2800, 0x2C00, 0x3000, 0x3400, 0x3800, 0x3C00,
0x2080, 0x2480, 0x2880, 0x2C80, 0x3080, 0x3480, 0x3880, 0x3C80,
0x2100, 0x2500, 0x2900, 0x2D00, 0x3100, 0x3500, 0x3900, 0x3D00,
0x2180, 0x2580, 0x2980, 0x2D80, 0x3180, 0x3580, 0x3980, 0x3D80,
0x2200, 0x2600, 0x2A00, 0x2E00, 0x3200, 0x3600, 0x3A00, 0x3E00,
0x2280, 0x2680, 0x2A80, 0x2E80, 0x3280, 0x3680, 0x3A80, 0x3E80,
0x2300, 0x2700, 0x2B00, 0x2F00, 0x3300, 0x3700, 0x3B00, 0x3F00,
0x2380, 0x2780, 0x2B80, 0x2F80, 0x3380, 0x3780, 0x3B80, 0x3F80,
0x2028, 0x2428, 0x2828, 0x2C28, 0x3028, 0x3428, 0x3828, 0x3C28,
0x20A8, 0x24A8, 0x28A8, 0x2CA8, 0x30A8, 0x34A8, 0x38A8, 0x3CA8,
0x2128, 0x2528, 0x2928, 0x2D28, 0x3128, 0x3528, 0x3928, 0x3D28,
0x21A8, 0x25A8, 0x29A8, 0x2DA8, 0x31A8, 0x35A8, 0x39A8, 0x3DA8,
0x2228, 0x2628, 0x2A28, 0x2E28, 0x3228, 0x3628, 0x3A28, 0x3E28,
0x22A8, 0x26A8, 0x2AA8, 0x2EA8, 0x32A8, 0x36A8, 0x3AA8, 0x3EA8,
0x2328, 0x2728, 0x2B28, 0x2F28, 0x3328, 0x3728, 0x3B28, 0x3F28,
0x23A8, 0x27A8, 0x2BA8, 0x2FA8, 0x33A8, 0x37A8, 0x3BA8, 0x3FA8,
0x2050, 0x2450, 0x2850, 0x2C50, 0x3050, 0x3450, 0x3850, 0x3C50,
0x20D0, 0x24D0, 0x28D0, 0x2CD0, 0x30D0, 0x34D0, 0x38D0, 0x3CD0,
0x2150, 0x2550, 0x2950, 0x2D50, 0x3150, 0x3550, 0x3950, 0x3D50,
0x21D0, 0x25D0, 0x29D0, 0x2DD0, 0x31D0, 0x35D0, 0x39D0, 0x3DD0,
0x2250, 0x2650, 0x2A50, 0x2E50, 0x3250, 0x3650, 0x3A50, 0x3E50,
0x22D0, 0x26D0, 0x2AD0, 0x2ED0, 0x32D0, 0x36D0, 0x3AD0, 0x3ED0,
0x2350, 0x2750, 0x2B50, 0x2F50, 0x3350, 0x3750, 0x3B50, 0x3F50,
0x23D0, 0x27D0, 0x2BD0, 0x2FD0, 0x33D0, 0x37D0, 0x3BD0, 0x3FD0};


unsigned char hrbuf[8192];

/* read input file */
int read_bin(unsigned char *name)
{

	FILE *fp;
	fp = fopen(name,"rb");
	if (NULL == fp)return INVALID;
	fread(hrbuf,1,8192,fp);
	fclose(fp);

	return SUCCESS;

}


/* pcx encoder helper */
/* the writer for the encline function */
int encput(unsigned char byt,unsigned char cnt, FILE *pcxfile)
{
          if(cnt){
            if((cnt==1)&& (0xc0 != (0xc0 &byt))){
                if(EOF == fputc((int)byt,pcxfile))
                return(0);
                return(1);
            }
            else{
                if(EOF==fputc((int)0xc0|cnt,pcxfile))
                return(0);
                if(EOF==fputc((int)byt,pcxfile))
                return(0);
                return(2);
            }
        }
        return(0);

}

/* encode a line in pcxformat encoding */
/* encodes a raw line and writes it out to disk */
int encline(unsigned char *inbuff,int inlen, FILE *pcxfile)
{
    unsigned char this,last;
    int srcindex,i;
    int total;
    unsigned char runcount;
    total=0;
    last = *(inbuff); runcount=1;

for(srcindex=1;srcindex!=inlen;srcindex++){
    this= *(++inbuff);
    if(this==last){
        runcount++;
        if(runcount==63){
            if(!(i=encput(last,runcount,pcxfile)))
            return(0);
            total+=i;
            runcount=0;
        }
    }
    else{
        if(runcount){
            if(!(i=encput(last,runcount,pcxfile)))
            return(0);
            total+=i;
        }
        last=this;
        runcount=1;
      }
   }

   if(runcount){
    if(!(i=encput(last,runcount,pcxfile)))
    return(0);
    return(total+i);
    }
    return (total);

}

unsigned char bigbuf[7680];

int makerag(FILE *fp2, FILE *fp3)
{
    int x, y, xoff, cnt = 0;
    unsigned char *ptr;

    for (y = 0; y < 192; y++) {
	    xoff = HB[y] - 0x2000;
    	ptr  = (unsigned char *) &hrbuf[xoff];
    	for (x=0;x<40;x++) {
		    bigbuf[cnt] = ptr[x];
		    cnt++;
		}
		fwrite(ptr,1,40,fp2);
		if (y < 160) fwrite(ptr,1,40,fp3);

	}
    return 0;

}



int save_to_X(unsigned char *basename)
{

    FILE *fp,*fp2,*fp3;
    unsigned char raxfile[128], botfile[128], topfile[128];

    sprintf(raxfile,"%s.RAX",basename);
    sprintf(botfile,"%s.BOT",basename);
    sprintf(topfile,"%s.TOP",basename);

 	fp = fopen(raxfile,"wb");
	if (NULL == fp)return INVALID;

	fp2 = fopen(botfile,"wb");
	if (NULL == fp2) {
		fclose(fp);
		remove(raxfile);
		return INVALID;
	}

    fp3 = fopen(topfile,"wb");
	if (NULL == fp3) {
		fclose(fp);
		fclose(fp2);
		remove(raxfile);
		remove(botfile);
		return INVALID;
	}

    /* width in bytes x height in rasters */
    fputc(40,fp);	 fputc(40,fp2);	   fputc(40,fp3);
    fputc(192,fp);	 fputc(192,fp2);   fputc(160,fp3);

	makerag(fp2,fp3);
	fclose(fp2);
	fclose(fp3);
    encline(bigbuf,7680,fp);
    fclose(fp);

    return SUCCESS;

}


void main(int argc, char **argv)
{
  int status = 0;

  unsigned char fname[128],infile[128],outfile[128];
  unsigned char *wordptr;

  if(argc == 1) {
    puts(szTextTitle);
    puts("Command line Usage is \"RAXPACK MyHires.BIN\"");
    puts("                      \"RAXPACK MyHires.BIN OutfileBaseName\"");
    printf("Enter Input FileName (Blank to Exit): ");
    gets(fname);
    if (fname[0] == ASCIIZ)
      exit(1);
    printf("Enter Output FileBaseName (Blank for Same) : ");
    gets(outfile);
  }
  else {
    strcpy(fname, argv[1]);
    if (argc > 2)
      strcpy(outfile, argv[2]);
    else
      outfile[0] = ASCIIZ;
  }

  strcpy(infile, fname);
  wordptr = strtok(infile, ".");
  if (outfile[0] == ASCIIZ)strcpy(outfile,infile);

  status = read_bin(fname);
  if (status) {
    puts(szTextTitle);
    printf("%s is an Unsupported Format or cannot be opened.\n", fname);
    exit(1);
  }

 status = save_to_X(outfile);
 if (status == SUCCESS)
	 printf("%s.RAX and %s.BOT and .TOP Saved!\n", outfile,outfile);
 else
    printf("Error saving %s.RAX and %s.BOT and .TOP!\n",outfile,outfile);

 puts("Have a Nice Dos!");
 exit(0);

}
